# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

from bisect import bisect_left, bisect_right
from datetime import datetime
import logging
logger = logging.getLogger(__name__)

import numpy as np
from netCDF4 import num2date
import xarray as xr

from opendrift.readers.basereader import BaseReader, vector_pairs_xy, StructuredReader
from opendrift.readers.roppy import depth


class Reader(BaseReader, StructuredReader):

    def __init__(self, filename=None, name=None, gridfile=None, standard_name_mapping={}):

        if filename is None:
            raise ValueError('Need filename as argument to constructor')

        # Map ROMS variable names to CF standard_name
        self.ROMS_variable_mapping = {
            # Removing (temoprarily) land_binary_mask from ROMS-variables,
            # as this leads to trouble with linearNDFast interpolation
            'mask_rho': 'land_binary_mask',
            'mask_psi': 'land_binary_mask',
            'h': 'sea_floor_depth_below_sea_level',
            'zeta': 'sea_surface_height',
            'u': 'x_sea_water_velocity',
            'v': 'y_sea_water_velocity',
            'u_eastward': 'x_sea_water_velocity',
            'v_northward': 'y_sea_water_velocity',
            'w': 'upward_sea_water_velocity',
            'temp': 'sea_water_temperature',
            'salt': 'sea_water_salinity',
            'uice': 'sea_ice_x_velocity',
            'vice': 'sea_ice_y_velocity',
            'aice': 'sea_ice_area_fraction',
            'hice': 'sea_ice_thickness',
            'gls': 'turbulent_generic_length_scale',
            'tke': 'turbulent_kinetic_energy',
            'AKs': 'ocean_vertical_diffusivity',
            'sustr': 'surface_downward_x_stress',
            'svstr': 'surface_downward_y_stress',
            'Uwind': 'x_wind',
            'Vwind': 'y_wind'}

        # Add user provided variable mappings
        self.ROMS_variable_mapping.update(standard_name_mapping)

        # z-levels to which sigma-layers may be interpolated
        self.zlevels = np.array([
            0, -.5, -1, -3, -5, -10, -25, -50, -75, -100, -150, -200,
            -250, -300, -400, -500, -600, -700, -800, -900, -1000, -1500,
            -2000, -2500, -3000, -3500, -4000, -4500, -5000, -5500, -6000,
            -6500, -7000, -7500, -8000])

        gls_param = ['gls_cmu0', 'gls_p', 'gls_m', 'gls_n']

        filestr = str(filename)
        if name is None:
            self.name = filestr
        else:
            self.name = name

        try:
            # Open file, check that everything is ok
            logger.info('Opening dataset: ' + filestr)
            if ('*' in filestr) or ('?' in filestr) or ('[' in filestr):
                logger.info('Opening files with MFDataset')
                def drop_non_essential_vars_pop(ds):
                    dropvars = [v for v in ds.variables if v not in
                                list(self.ROMS_variable_mapping.keys()) + gls_param +
                                ['ocean_time', 's_rho', 'Cs_r', 'hc', 'angle']
                                and v[0:3] not in ['lon', 'lat', 'mas']]
                    logger.debug('Dropping variables: %s' % dropvars)
                    ds = ds.drop_vars(dropvars)
                    return ds
                self.Dataset = xr.open_mfdataset(filename,
                    chunks={'ocean_time': 1}, compat='override', decode_times=False,
                    preprocess=drop_non_essential_vars_pop,
                    data_vars='minimal', coords='minimal')
            else:
                logger.info('Opening file with Dataset')
                self.Dataset = xr.open_dataset(filename, decode_times=False)
        except Exception as e:
            raise ValueError(e)


        if 'Vtransform' in self.Dataset.variables:
            self.Vtransform = self.Dataset.variables['Vtransform'].data  # scalar
        else:
            logger.warning('Vtransform not found, using 1')
            self.Vtransform = 1
        self.Vtransform = np.asarray(self.Vtransform)

        if 's_rho' not in self.Dataset.variables:
            dimensions = 2
        else:
            dimensions = 3

        if dimensions == 3:
            # Read sigma-coordinate values
            try:
                self.sigma = self.Dataset.variables['s_rho'][:]
                self.sigma = self.sigma.data  # Extract data
            except:
                num_sigma = len(self.Dataset.dimensions['s_rho'])
                logger.warning(
                    's_rho not available in dataset, constructing from'
                    ' number of layers (%s).' % num_sigma)
                self.sigma = (np.arange(num_sigma)+.5-num_sigma)/num_sigma

            # Read sigma-coordinate transform parameters
            try:
                self.Dataset.variables['Cs_r'].set_auto_mask(False)
            except:
                pass
            self.Cs_r = self.Dataset.variables['Cs_r'][:]
            try:
                self.hc = self.Dataset.variables['hc'][:]
            except:
                self.hc = self.Dataset.variables['hc'].data  # scalar

            self.num_layers = len(self.sigma)
        else:
            self.num_layers = 1
            self.ROMS_variable_mapping['ubar'] = 'x_sea_water_velocity'
            self.ROMS_variable_mapping['vbar'] = 'y_sea_water_velocity'
            del self.ROMS_variable_mapping['u']
            del self.ROMS_variable_mapping['v']

        for var in list(self.ROMS_variable_mapping):  # Remove unused variables
            if var not in self.Dataset.variables:
                del self.ROMS_variable_mapping[var]

        if 'lat_rho' in self.Dataset.variables:
            # Horizontal oordinates and directions
            self.lat = self.Dataset.variables['lat_rho'][:]
            self.lon = self.Dataset.variables['lon_rho'][:]
            self.lon = self.lon.data  # Extract, could be avoided downstream
            self.lat = self.lat.data
            if self.lat.ndim == 1:
                self.lon, self.lat = np.meshgrid(self.lon, self.lat)
                self.angle_xi_east = 0
        else:
            if gridfile is None:
                raise ValueError(filename + ' does not contain lon/lat '
                                 'arrays, please supply a grid-file '
                                 '"gridfile=<grid_file>"')
            else:
                gf = xr.open_dataset(gridfile)
                gridvars = ['lat_rho', 'notvar', 'lon_rho', 'mask_rho', 'mask_u', 'mask_v', 'angle', 'h']
                for gv in gridvars:
                    if gv in gf.variables:
                        if gv in ['lat_rho', 'lon_rho']:
                            setname = gv[0:3]
                            setattr(self, setname, gf.variables[gv][:].data)
                        elif gv in ['angle']:
                            setname = 'angle_xi_east'
                            setattr(self, setname, gf.variables[gv][:])
                        else:
                            setname = gv
                            setattr(self, gv, gf.variables[gv][:])
                if self.lat.ndim == 1:
                    self.lon, self.lat = np.meshgrid(self.lon, self.lat)
                    self.angle_xi_east = 0

        try:  # Check for GLS parameters (diffusivity)
            self.gls_parameters = {}
            for gls_par in gls_param:
                self.gls_parameters[gls_par] = \
                    self.Dataset.variables[gls_par][()]
            logger.info('Read GLS parameters from file.')
        except Exception as e:
            logger.info(e)
            logger.info('Did not find complete set of GLS parameters')

        # Get time coverage
        try:
            ocean_time = self.Dataset.variables['ocean_time']
        except:
            ocean_time = self.Dataset.variables['time']
        time_units = ocean_time.attrs['units']
        if time_units == 'second':
            logger.info('Ocean time given as seconds relative to start '
                         'Setting artifical start time of 1 Jan 2000.')
            time_units = 'seconds since 2000-01-01 00:00:00'
        self.times = num2date(ocean_time[:], time_units)
        self.start_time = self.times[0]
        self.end_time = self.times[-1]
        if len(self.times) > 1:
            self.time_step = self.times[1] - self.times[0]
        else:
            self.time_step = None

        self.name = 'roms native'

        self.precalculate_s2z_coefficients = True

        # Find all variables having standard_name
        self.variables = []
        for var_name in self.Dataset.variables:
            var = self.Dataset.variables[var_name]
            if 'standard_name' in var.attrs and var_name not in self.ROMS_variable_mapping.keys():
                self.ROMS_variable_mapping[var_name] = var.attrs['standard_name']
            if var_name in self.ROMS_variable_mapping.keys():
                self.variables.append(self.ROMS_variable_mapping[var_name])

        # A bit hackish solution:
        # If variable names or their standard_name contain "east" or "north", 
        # these should not be rotated from xi-direction to east-direction
        self.do_not_rotate = []
        for var, stdname in self.ROMS_variable_mapping.items():
            if 'east' in var.lower() or 'east' in stdname.lower() or \
                    'north' in var.lower() or 'north' in stdname.lower():
                self.do_not_rotate.append(stdname)
        if len(self.do_not_rotate)>0:
            logger.debug('The following ROMS vectors are considered east-north, and will not be rotated %s' % self.do_not_rotate)

        # Run constructor of parent Reader class
        super(Reader, self).__init__()

    def get_variables(self, requested_variables, time=None,
                      x=None, y=None, z=None):
        start_time = datetime.now()
        requested_variables, time, x, y, z, outside = self.check_arguments(
            requested_variables, time, x, y, z)

        # If one vector component is requested, but not the other
        # we must add the other for correct rotation
        for vector_pair in vector_pairs_xy:
            if (vector_pair[0] in requested_variables and
                vector_pair[1] not in requested_variables):
                requested_variables.extend([vector_pair[1]])
            if (vector_pair[1] in requested_variables and
                vector_pair[0] not in requested_variables):
                requested_variables.extend([vector_pair[0]])

        nearestTime, dummy1, dummy2, indxTime, dummy3, dummy4 = \
            self.nearest_time(time)

        variables = {}

        if z is None:
            z = np.atleast_1d(0)

        # Find horizontal indices corresponding to requested x and y
        if hasattr(self, 'clipped'):
            clipped = self.clipped
        else: clipped = 0
        indx = np.floor((x-self.xmin)/self.delta_x).astype(int) + clipped
        indy = np.floor((y-self.ymin)/self.delta_y).astype(int) + clipped
        indx[outside] = 0  # To be masked later
        indy[outside] = 0
        indx_el = indx.copy()
        indy_el = indy.copy()
        # Adding buffer, to cover also future positions of elements
        buffer = self.buffer
        # Avoiding the last pixel in each dimension, since there are
        # several grids which are shifted (rho, u, v, psi)
        indx = np.arange(np.max([0, indx.min()-buffer]),
                            np.min([indx.max()+buffer, self.lon.shape[1]-1]))
        indy = np.arange(np.max([0, indy.min()-buffer]),
                            np.min([indy.max()+buffer, self.lon.shape[0]-1]))

        # Find depth levels covering all elements
        if z.min() == 0 or not hasattr(self, 'hc'):
            indz = self.num_layers - 1  # surface layer
            variables['z'] = 0

        else:
            # Find the range of indices covering given z-values
            if not hasattr(self, 'sea_floor_depth_below_sea_level'):
                logger.debug('Reading sea floor depth...')
                self.sea_floor_depth_below_sea_level = \
                    self.Dataset.variables['h'][:]

                Htot = self.sea_floor_depth_below_sea_level
                self.z_rho_tot = depth.sdepth(Htot, self.hc, self.Cs_r,
                                              Vtransform=self.Vtransform)

            H = self.sea_floor_depth_below_sea_level[indy, indx]
            z_rho = depth.sdepth(H, self.hc, self.Cs_r,
                                 Vtransform=self.Vtransform)
            # Element indices must be relative to extracted subset
            indx_el = np.clip(indx_el - indx.min(), 0, z_rho.shape[2]-1)
            indy_el = np.clip(indy_el - indy.min(), 0, z_rho.shape[1]-1)

            # Loop to find the layers covering the requested z-values
            indz_min = 0
            indz_max = self.num_layers
            for i in range(self.num_layers):
                if np.min(z-z_rho[i, indy_el, indx_el]) > 0:
                    indz_min = i
                if np.max(z-z_rho[i, indy_el, indx_el]) > 0:
                    indz_max = i
            indz = range(np.maximum(0, indz_min-self.verticalbuffer),
                         np.minimum(self.num_layers,
                                    indz_max + 1 + self.verticalbuffer))
            z_rho = z_rho[indz, :, :]
            # Determine the z-levels to which to interpolate
            zi1 = np.maximum(0, bisect_left(-np.array(self.zlevels),
                                            -z.max()) - self.verticalbuffer)
            zi2 = np.minimum(len(self.zlevels),
                             bisect_right(-np.array(self.zlevels),
                                          -z.min()) + self.verticalbuffer)
            variables['z'] = np.array(self.zlevels[zi1:zi2])

        #read_masks = {}  # To store maskes for various grids
        mask_values = {}
        for par in requested_variables:
            varname = [name for name, cf in
                       self.ROMS_variable_mapping.items() if cf == par]
            var = self.Dataset.variables[varname[0]]

            if par == 'land_binary_mask':
                if not hasattr(self, 'land_binary_mask'):
                    # Read landmask for whole domain, for later re-use
                    self.land_binary_mask = \
                        1 - self.Dataset.variables['mask_rho'][:]
                variables[par] = self.land_binary_mask[indy, indx]
            elif var.ndim == 2:
                variables[par] = var[indy, indx]
            elif var.ndim == 3:
                variables[par] = var[indxTime, indy, indx]
            elif var.ndim == 4:
                variables[par] = var[indxTime, indz, indy, indx]
            else:
                raise Exception('Wrong dimension of variable: ' +
                                self.variable_mapping[par])

            variables[par] = np.asarray(variables[par])  # If Xarray
            start = datetime.now()

            if par not in mask_values:
                indxgrid = indx
                indygrid = indy
                if par in ['x_sea_water_velocity', 'sea_water_x_velocity',
                           'eastward_sea_water_velocity']:
                    if not hasattr(self, 'mask_u'):
                        if 'mask_u' in self.Dataset.variables:
                            self.mask_u = self.Dataset.variables['mask_u'][:]
                        elif 'mask_rho' in self.Dataset.variables:
                            self.mask_u = self.Dataset.variables['mask_rho'][:]
                        else:
                            continue
                    mask = self.mask_u[indygrid, indxgrid]
                elif par in ['y_sea_water_velocity', 'sea_water_y_velocity',
                             'northward_sea_water_velocity']:
                    if not hasattr(self, 'mask_v'):
                        if 'mask_v' in self.Dataset.variables:
                            self.mask_v = self.Dataset.variables['mask_v'][:]
                        elif 'mask_rho' in self.Dataset.variables:
                            self.mask_v = self.Dataset.variables['mask_rho'][:]
                        else:
                            continue
                    mask = self.mask_v[indygrid, indxgrid]
                else:
                    if not hasattr(self, 'mask_rho'):
                        # For ROMS-Agrif this must perhaps be mask_psi?
                        self.mask_rho = self.Dataset.variables['mask_rho'][:]
                    mask = self.mask_rho[indygrid, indxgrid]
                mask = np.asarray(mask)
                if mask.min() == 0 and par != 'land_binary_mask':
                    first_mask_point = np.where(mask.ravel()==0)[0][0]
                    if variables[par].ndim == 3:
                        upper = variables[par][0,:,:]
                    else:
                        upper = variables[par]
                    mask_values[par] = upper.ravel()[first_mask_point]
                    variables[par][variables[par]==mask_values[par]] = np.nan

            if var.ndim == 4:
                # Regrid from sigma to z levels
                if len(np.atleast_1d(indz)) > 1:
                    logger.debug('sigma to z for ' + varname[0])
                    if self.precalculate_s2z_coefficients is True:
                        M = self.sea_floor_depth_below_sea_level.shape[0]
                        N = self.sea_floor_depth_below_sea_level.shape[1]
                        O = len(self.z_rho_tot)
                        if not hasattr(self, 's2z_A'):
                            logger.debug('Calculating sigma2z-coefficients for whole domain')
                            starttime = datetime.now()
                            dummyvar = np.ones((O, M, N))
                            dummy, self.s2z_total = depth.multi_zslice(dummyvar, self.z_rho_tot, self.zlevels)
                            # Store arrays/coefficients
                            self.s2z_A = self.s2z_total[0].reshape(len(self.zlevels), M, N)
                            self.s2z_C = self.s2z_total[1].reshape(len(self.zlevels), M, N)
                            #self.s2z_I = self.s2z_total[2].reshape(M, N)
                            self.s2z_kmax = self.s2z_total[3]
                            del self.s2z_total  # Free memory
                            logger.info('Time: ' + str(datetime.now() - starttime))
                        if 'A' not in locals():
                            logger.debug('Re-using sigma2z-coefficients')
                            # Select relevant subset of full arrays
                            zle = np.arange(zi1, zi2)  # The relevant depth levels
                            A = self.s2z_A.copy()  # Awkward subsetting to prevent losing one dimension
                            A = A[:,:,indx]
                            A = A[:,indy,:]
                            A = A[zle,:,:]
                            C = self.s2z_C.copy()
                            C = C[:,:,indx]
                            C = C[:,indy,:]
                            C = C[zle,:,:]
                            C = C - C.max() + variables[par].shape[0] - 1
                            C[C<1] = 1
                            A = A.reshape(len(zle), len(indx)*len(indy))
                            C = C.reshape(len(zle), len(indx)*len(indy))
                            I = np.arange(len(indx)*len(indy))
                            ## Check
                            #dummyvar2, s2z = depth.multi_zslice(
                            #    variables[par].copy(), z_rho.copy(), variables['z'])
                            #print len(zle), variables[par].shape, 'zle, varshape'
                            #Ac,Cc,Ic,kmaxc = s2z
                            #print C, 'C'
                            #print Cc, 'Cc'
                            #print C.shape, Cc.shape
                            #if C.max() != Cc.max():
                            #    print 'WARNING!!'
                            #    import sys; sys.exit('stop')
                            kmax = len(zle)  # Must be checked. Or number of sigma-layers?
                    if 'A' not in locals():
                        logger.debug('Calculating new sigma2z-coefficients')
                        variables[par], s2z = depth.multi_zslice(
                            variables[par], z_rho, variables['z'])
                        A,C,I,kmax = s2z
                        # Reshaping to compare with subset of full array
                        #zle = np.arange(zi1, zi2)
                        #A = A.reshape(len(zle), len(indx), len(indy))
                        #C = C.reshape(len(zle), len(indx), len(indy))
                        #I = I.reshape(len(indx), len(indy))
                    else:
                        logger.debug('Applying sigma2z-coefficients')
                        # Re-using sigma2z koefficients:
                        F = np.asarray(variables[par])
                        Fshape = F.shape
                        N = F.shape[0]
                        M = F.size // N
                        F = F.reshape((N, M))
                        R = (1-A)*F[(C-1, I)]+A*F[(C, I)]
                        variables[par] = R.reshape((kmax,) + Fshape[1:])

                    # Nan in input to multi_zslice gives extreme values in output
                    variables[par][variables[par]>1e+9] = np.nan

            # Mask values outside domain
            variables[par] = np.ma.array(variables[par], ndmin=2, mask=False)

            # Skipping de-staggering, as it leads to invalid values at later interpolation
            #if block is True:
            #    # Unstagger grid for vectors
            #    logger.debug('Unstaggering ' + par)
            #    if 'eta_v' in var.dimensions:
            #        variables[par] = np.ma.array(variables[par],
            #                            mask=variables[par].mask)
            #        variables[par][variables[par].mask] = 0
            #        if variables[par].ndim == 2:
            #            variables[par] = \
            #                (variables[par][0:-1,0:-1] +
            #                variables[par][0:-1,1::])/2
            #        elif variables[par].ndim == 3:
            #            variables[par] = \
            #                (variables[par][:,0:-1,0:-1] +
            #                variables[par][:,0:-1,1::])/2
            #        variables[par] = np.ma.masked_where(variables[par]==0,
            #                                            variables[par])
            #    elif 'eta_u' in var.dimensions:
            #        variables[par] = np.ma.array(variables[par],
            #                            mask=variables[par].mask)
            #        variables[par][variables[par].mask] = 0
            #        if variables[par].ndim == 2:
            #            variables[par] = \
            #                (variables[par][0:-1,0:-1] +
            #                 variables[par][1::,0:-1])/2
            #        elif variables[par].ndim == 3:
            #            variables[par] = \
            #                (variables[par][:,0:-1,0:-1] +
            #                 variables[par][:,1::,0:-1])/2
            #        variables[par] = np.ma.masked_where(variables[par]==0,
            #                                            variables[par])
            #    else:
            #        if variables[par].ndim == 2:
            #            variables[par] = variables[par][1::, 1::]
            #        elif variables[par].ndim == 3:
            #            variables[par] = variables[par][:,1::, 1::]

        # TODO: should be midpoints, but angle array below needs integer
        #indx = indx[0:-1]  # Only if de-staggering has been performed
        #indy = indy[1::]
        variables['x'] = indx
        variables['y'] = indy

        variables['x'] = variables['x'].astype(np.float32)
        variables['y'] = variables['y'].astype(np.float32)
        variables['time'] = nearestTime

        if 'x_sea_water_velocity' in variables.keys() or \
            'sea_ice_x_velocity' in variables.keys() or \
            'x_wind' in variables.keys():
            # We must rotate current vectors
            if not hasattr(self, 'angle_xi_east'):
                if 'angle' in self.Dataset.variables:
                    logger.debug('Reading angle between xi and east...')
                    self.angle_xi_east = self.Dataset.variables['angle'][:]
            if isinstance(self.angle_xi_east, int):
                rad = self.angle_xi_east
            else:
                rad = self.angle_xi_east[indy, indx]
                rad = np.ma.asarray(rad)
            if 'x_sea_water_velocity' in variables.keys() and \
                    'x_sea_water_velocity' not in self.do_not_rotate:
                variables['x_sea_water_velocity'], \
                    variables['y_sea_water_velocity'] = rotate_vectors_angle(
                        variables['x_sea_water_velocity'],
                        variables['y_sea_water_velocity'], rad)
            if 'sea_ice_x_velocity' in variables.keys() and \
                    'sea_ice_x_velocity' not in self.do_not_rotate:
                variables['sea_ice_x_velocity'], \
                    variables['sea_ice_y_velocity'] = rotate_vectors_angle(
                        variables['sea_ice_x_velocity'],
                        variables['sea_ice_y_velocity'], rad)
            if 'x_wind' in variables.keys() and \
                    'x_wind' not in self.do_not_rotate:
                variables['x_wind'], \
                    variables['y_wind'] = rotate_vectors_angle(
                        variables['x_wind'],
                        variables['y_wind'], rad)

        # Masking NaN
        for var in requested_variables:
            variables[var] = np.ma.masked_invalid(variables[var])

        logger.debug('Time for ROMS native reader: ' + str(datetime.now()-start_time))

        return variables


def rotate_vectors_angle(u, v, radians):
    u2 = u*np.cos(radians) - v*np.sin(radians)
    v2 = u*np.sin(radians) + v*np.cos(radians)
    return u2, v2



"""
dataclass to hold the CCME data

CCME is kind of a special case, so this nails it down
"""
from dataclasses import dataclass, field

from ..common.utilities import dataclass_to_json

from ..common.measurement import MassFraction


@dataclass_to_json
@dataclass
class CCME:
    """
    hold the CCME data from env CA

    (https://www.ccme.ca/en/resources/canadian_environmental_quality_guidelines/index.html)
    """
    F1: MassFraction = None
    F2: MassFraction = None
    F3: MassFraction = None
    F4: MassFraction = None
    method: str = None





# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2021, Trond Kristiansen, Niva
# Jan 2021 Simplified by Knut-Frode Dagestad, MET Norway, and adapted to to Kvile et al. (2018)

import datetime
import numpy as np
import logging; logger = logging.getLogger(__name__)
from opendrift.models.oceandrift import Lagrangian3DArray, OceanDrift


class LarvalFishElement(Lagrangian3DArray):
    """
    Extending Lagrangian3DArray with specific properties for larval and juvenile stages of fish
    """

    variables = Lagrangian3DArray.add_variables([
        ('diameter', {'dtype': np.float32,
                      'units': 'm',
                      'default': 0.0014}),  # for NEA Cod
        ('neutral_buoyancy_salinity', {'dtype': np.float32,
                                       'units': 'PSU',
                                       'default': 31.25}),  # for NEA Cod
        ('stage_fraction', {'dtype': np.float32,  # to track percentage of development time completed
                            'units': '',
                            'default': 0.}),
        ('hatched', {'dtype': np.uint8,  # 0 for eggs, 1 for larvae
                     'units': '',
                     'default': 0}),
        ('length', {'dtype': np.float32,
                    'units': 'mm',
                    'default': 0}),
        ('weight', {'dtype': np.float32,
                    'units': 'mg',
                    'default': 0.08}),
        ('survival', {'dtype': np.float32,  # Not yet used
                      'units': '',
                      'default': 1.})])


class LarvalFish(OceanDrift):
    """Buoyant particle trajectory model based on the OpenDrift framework.

        Developed at MET Norway

        Generic module for particles that are subject to vertical turbulent
        mixing with the possibility for positive or negative buoyancy

        Particles could be e.g. oil droplets, plankton, or sediments

    """

    ElementType = LarvalFishElement

    required_variables = {
        'x_sea_water_velocity': {'fallback': 0},
        'y_sea_water_velocity': {'fallback': 0},
        'sea_surface_wave_significant_height': {'fallback': 0},
        'x_wind': {'fallback': 0},
        'y_wind': {'fallback': 0},
        'land_binary_mask': {'fallback': None},
        'sea_floor_depth_below_sea_level': {'fallback': 100},
        'ocean_vertical_diffusivity': {'fallback': 0.01, 'profiles': True},
        'ocean_mixed_layer_thickness': {'fallback': 50},
        'sea_water_temperature': {'fallback': 10, 'profiles': True},
        'sea_water_salinity': {'fallback': 34, 'profiles': True}
    }

    required_profiles_z_range = [0, -50]  # The depth range (in m) which profiles should cover

    def __init__(self, *args, **kwargs):

        # Calling general constructor of parent class
        super(LarvalFish, self).__init__(*args, **kwargs)

        # IBM configuration options
        self._add_config({
            'IBM:fraction_of_timestep_swimming':
                {'type': 'float', 'default': 0.15,
                 'min': 0.0, 'max': 1.0, 'units': 'fraction',
                 'description': 'Fraction of timestep swimming',
                 'level': self.CONFIG_LEVEL_ADVANCED},
            })

        self._set_config_default('drift:vertical_mixing', True)

    def update_terminal_velocity(self, Tprofiles=None,
                                 Sprofiles=None, z_index=None):
        """Calculate terminal velocity for Pelagic Egg

        according to
        S. Sundby (1983): A one-dimensional model for the vertical
        distribution of pelagic fish eggs in the mixed layer
        Deep Sea Research (30) pp. 645-661

        Method copied from ibm.f90 module of LADIM:
        Vikebo, F., S. Sundby, B. Aadlandsvik and O. Otteraa (2007),
        Fish. Oceanogr. (16) pp. 216-228
        """
        g = 9.81  # ms-2

        # Pelagic Egg properties that determine buoyancy
        eggsize = self.elements.diameter  # 0.0014 for NEA Cod
        eggsalinity = self.elements.neutral_buoyancy_salinity
        # 31.25 for NEA Cod

        # prepare interpolation of temp, salt
        if not (Tprofiles is None and Sprofiles is None):
            if z_index is None:
                z_i = range(Tprofiles.shape[0])  # evtl. move out of loop
                # evtl. move out of loop
                z_index = interp1d(-self.environment_profiles['z'],
                                   z_i, bounds_error=False)
            zi = z_index(-self.elements.z)
            upper = np.maximum(np.floor(zi).astype(np.uint8), 0)
            lower = np.minimum(upper + 1, Tprofiles.shape[0] - 1)
            weight_upper = 1 - (zi - upper)

        # do interpolation of temp, salt if profiles were passed into
        # this function, if not, use reader by calling self.environment
        if Tprofiles is None:
            T0 = self.environment.sea_water_temperature
        else:
            T0 = Tprofiles[upper, range(Tprofiles.shape[1])] * \
                 weight_upper + \
                 Tprofiles[lower, range(Tprofiles.shape[1])] * \
                 (1 - weight_upper)
        if Sprofiles is None:
            S0 = self.environment.sea_water_salinity
        else:
            S0 = Sprofiles[upper, range(Sprofiles.shape[1])] * \
                 weight_upper + \
                 Sprofiles[lower, range(Sprofiles.shape[1])] * \
                 (1 - weight_upper)

        # The density difference between a pelagic egg and the ambient water
        # is regulated by their salinity difference through the
        # equation of state for sea water.
        # The Egg has the same temperature as the ambient water and its
        # salinity is regulated by osmosis through the egg shell.
        DENSw = self.sea_water_density(T=T0, S=S0)
        DENSegg = self.sea_water_density(T=T0, S=eggsalinity)
        dr = DENSw - DENSegg  # density difference

        # water viscosity
        my_w = 0.001 * (1.7915 - 0.0538 * T0 + 0.007 * (T0 ** (2.0)) - 0.0023 * S0)
        # ~0.0014 kg m-1 s-1

        # terminal velocity for low Reynolds numbers
        W = (1.0 / my_w) * (1.0 / 18.0) * g * eggsize ** 2 * dr

        # check if we are in a Reynolds regime where Re > 0.5
        highRe = np.where(W * 1000 * eggsize / my_w > 0.5)

        # Use empirical equations for terminal velocity in
        # high Reynolds numbers.
        # Empirical equations have length units in cm!
        my_w = 0.01854 * np.exp(-0.02783 * T0)  # in cm2/s
        d0 = (eggsize * 100) - 0.4 * \
             (9.0 * my_w ** 2 / (100 * g) * DENSw / dr) ** (1.0 / 3.0)  # cm
        W2 = 19.0 * d0 * (0.001 * dr) ** (2.0 / 3.0) * (my_w * 0.001 * DENSw) ** (-1.0 / 3.0)
        # cm/s
        W2 = W2 / 100.  # back to m/s

        W[highRe] = W2[highRe]
        self.elements.terminal_velocity = W

    def fish_growth(self, weight, temperature):
        # Weight in milligrams, temperature in celcius
        # Daily growth rate in percent according to
        # Folkvord, A. 2005. "Comparison of Size-at-Age of Larval Atlantic Cod (Gadus Morhua)
        # from Different Populations Based on Size- and Temperature-Dependent Growth
        # Models." Canadian Journal of Fisheries and Aquatic Sciences.
        # Journal Canadien Des Sciences Halieutiques # et Aquatiques 62(5): 1037-52.
        GR = 1.08 + 1.79 * temperature - 0.074 * temperature * np.log(weight) \
             - 0.0965 * temperature * np.log(weight) ** 2 \
             + 0.0112 * temperature * np.log(weight) ** 3

        # Growth rate(g) converted to milligram weight (gr_mg) per timestep:
        g = (np.log(GR / 100. + 1)) * self.time_step.total_seconds()/86400  
        return weight * (np.exp(g) - 1.)

    def update_fish_larvae(self):

        # Hatching of eggs
        eggs = np.where(self.elements.hatched==0)[0]
        if len(eggs) > 0:
            amb_duration = np.exp(3.65 - 0.145*self.environment.sea_water_temperature[eggs]) # Total egg development time (days) according to ambient temperature (Ellertsen et al. 1988)
            days_in_timestep = self.time_step.total_seconds()/(60*60*24)  # The fraction of a day completed in one time step
            amb_fraction = days_in_timestep/amb_duration # Fraction of development time completed during present time step
            self.elements.stage_fraction[eggs] += amb_fraction # Add fraction completed during present timestep to cumulative fraction completed
            hatching = np.where(self.elements.stage_fraction[eggs]>=1)[0]
            if len(hatching) > 0:
                logger.debug('Hatching %s eggs' % len(hatching))
                self.elements.hatched[eggs[hatching]] = 1 # Eggs with total development time completed are hatched (1)

        larvae = np.where(self.elements.hatched==1)[0]
        if len(larvae) == 0:
            logger.debug('%s eggs, with maximum stage_fraction of %s (1 gives hatching)'
                         % (len(eggs), self.elements.stage_fraction[eggs].max()))
            return

        # Increasing weight of larvae
        avg_weight_before = self.elements.weight[larvae].mean()
        growth = self.fish_growth(self.elements.weight[larvae],
                                  self.environment.sea_water_temperature[larvae])
        self.elements.weight[larvae] += growth
        avg_weight_after = self.elements.weight[larvae].mean()
        logger.debug('Growing %s larve from average size %s to %s' %
              (len(larvae), avg_weight_before, avg_weight_after))
    
        # Increasing length of larvae, according to Folkvord (2005)
        w = self.elements.weight[larvae]
        self.elements.length[larvae] = np.exp(2.296 + 0.277 * np.log(w) - 0.005128 *np.log10(w)**2)

    def larvae_vertical_migration(self):

        larvae = np.where(self.elements.hatched==1)[0]
        if len(larvae) == 0:
            return

        # Vertical migration of Larvae
        # Swim function from Peck et al. 2006
        L = self.elements.length[larvae]
        swim_speed = (0.261*(L**(1.552*L**(-0.08))) - 5.289/L) / 1000
        f = self.get_config('IBM:fraction_of_timestep_swimming')
        max_migration_per_timestep = f*swim_speed*self.time_step.total_seconds()

        # Using here UTC hours. Should be changed to local solar time,
        # although a phase shift of some hours should not make much difference
        if self.time.hour < 12:
            direction = -1  # Swimming down when light is increasing
        else:
            direction = 1  # Swimming up when light is decreasing

        self.elements.z[larvae] = np.minimum(0, self.elements.z[larvae] + direction*max_migration_per_timestep)

    def update(self):

        self.update_fish_larvae()
        self.advect_ocean_current()
        self.update_terminal_velocity()
        self.vertical_mixing()
        self.larvae_vertical_migration()



"""
versions.py

code for dealing with different versions of the data model

"""
from .oil import ADIOS_DATA_MODEL_VERSION


def update_json(py_json):
    """
    updates JSON for an oil object from an older version to a newer one
    """

    cur_ver = ADIOS_DATA_MODEL_VERSION
    ver = py_json.get('adios_data_model_version')

    if ver is None:
        # assume it's the version from before we added a version
        ver = "0.10.0"
    ver = Version.from_py_json(ver)
    if ver == cur_ver:  # nothing to be done
        return py_json
    elif ver > cur_ver:
        raise VersionError(f"Version: {ver} is not supported by this version of Oil object")
    else:
        raise VersionError(f"updater not available for version: {ver}")



from dataclasses import dataclass, field
from ..common.utilities import dataclass_to_json


@dataclass_to_json
@dataclass
class Label:
    '''
        So Labels will be a collection terms that the user can use to
        narrow down the list of oils he/she is interested in.
    '''
    name: str
    product_type: list = field(default_factory=list)



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

from datetime import datetime
import pyproj
import numpy as np
from netCDF4 import num2date
import logging
logger = logging.getLogger(__name__)

from opendrift.readers.basereader import BaseReader, StructuredReader
import xarray as xr

def proj_from_CF_dict(c):

    # This method should be extended to other projections:
    # https://cfconventions.org/wkt-proj-4.html
    if not 'grid_mapping_name' in c:
        raise ValueError('grid_mapping not given in dictionary')
    gm = c['grid_mapping_name']

    if 'earth_radius' in c:
        earth_radius = c['earth_radius']
    else:
        earth_radius = 6371000.
    if gm == 'polar_stereographic':
        lon_0 = 0  # default, but dangerous
        for l0 in ['longitude_of_projection_origin',
                   'longitude_of_central_meridian',
                   'straight_vertical_longitude_from_pole']:
            if l0 in c:
                lon_0 = c[l0]
        if 'latitude_of_origin' in c:
            lat_ts = c['latitude_of_origin']
        else:
            lat_ts = c['latitude_of_projection_origin']
        if 'false_easting' in c:
            x0 = c['false_easting']
        else:
            x0 = 0  # dangerous?
        if 'false_northing' in c:
            y0 = c['false_northing']
        else:
            y0 = 0  # dangerous: is there a better default?
        if 'scale_factor_at_projection_origin' in c:
            k0 = c['scale_factor_at_projection_origin']
        else:
            k0 = 1.0
        proj4 = ('+proj={!s} +lat_0={!s} +lon_0={!s} +lat_ts={!s} '
                 '+k_0={!s} +x_0={!s} +y_0={!s} +units=m +a={!s} '
                 '+no_defs'.format('stere',
                                   c['latitude_of_projection_origin'],
                                   lon_0, lat_ts, k0, x0, y0, earth_radius)
                 )

    elif gm == 'rotated_latitude_longitude':
        proj4 = '+proj=ob_tran +o_proj=longlat +lon_0=%s +o_lat_p=%s +R=%s +no_defs' % (
                c['grid_north_pole_longitude']-180, c['grid_north_pole_latitude'], earth_radius)

    proj = pyproj.Proj(proj4)

    return proj4, proj


class Reader(StructuredReader, BaseReader):
    """
    A reader for `CF-compliant <https://cfconventions.org/>`_ netCDF files. It can take a single file, or a file pattern.

    Args:
        :param filename: A single netCDF file, or a pattern of files. The
                         netCDF file can also be an URL to an OPeNDAP server.
        :type filename: string, requiered.

        :param name: Name of reader
        :type name: string, optional

        :param proj4: PROJ.4 string describing projection of data.
        :type proj4: string, optional

    Example:

    .. code::

       from opendrift.readers.reader_netCDF_CF_generic import Reader
       r = Reader("arome_subset_16Nov2015.nc")

    Several files can be specified by using a pattern:

    .. code::

       from opendrift.readers.reader_netCDF_CF_generic import Reader
       r = Reader("*.nc")

    or an OPeNDAP URL can be used:

    .. code::

       from opendrift.readers.reader_netCDF_CF_generic import Reader
       r = Reader('https://thredds.met.no/thredds/dodsC/mepslatest/meps_lagged_6_h_latest_2_5km_latest.nc')


    """

    def __init__(self, filename=None, name=None, proj4=None, standard_name_mapping={}, ensemble_member=None):
        if filename is None:
            raise ValueError('Need filename as argument to constructor')

        filestr = str(filename)
        if name is None:
            self.name = filestr
        else:
            self.name = name

        try:
            # Open file, check that everything is ok
            logger.info('Opening dataset: ' + filestr)
            if ('*' in filestr) or ('?' in filestr) or ('[' in filestr):
                logger.info('Opening files with MFDataset')
                self.Dataset = xr.open_mfdataset(filename, data_vars='minimal', coords='minimal',
                                                 chunks={'time': 1}, decode_times=False)
            elif ensemble_member is not None:
                self.Dataset = xr.open_dataset(filename, decode_times=False).isel(ensemble_member=ensemble_member)
            else:
                self.Dataset = xr.open_dataset(filename, decode_times=False)
        except Exception as e:
            raise ValueError(e)

        # NB: check below might not be waterproof
        if 'ocean_time' in self.Dataset.dims and 'eta_u' in self.Dataset.dims and \
                'eta_rho' in self.Dataset.dims:
            raise ValueError('This seems to be a ROMS native file, should use ROMS native reader instead')

        logger.debug('Finding coordinate variables.')
        if proj4 is not None:  # If user has provided a projection apriori
            self.proj4 = proj4
        # Find x, y and z coordinates
        lon_var_name = None
        lat_var_name = None
        self.unitfactor = 1
        self.realizations = None
        for var_name in self.Dataset.variables:
            var = self.Dataset.variables[var_name]

            if self.proj4 is None:
                if 'proj4' in var.attrs:
                    self.proj4 = str(var.attrs['proj4'])
                elif 'proj4_string' in var.attrs:
                    self.proj4 = str(var.attrs['proj4_string'])
                elif 'grid_mapping_name' in var.attrs:
                    logger.debug(
                        ('Parsing CF grid mapping dictionary:'
                        ' ' + str(var.attrs)))
                    try:
                        self.proj4, proj =\
                            proj_from_CF_dict(var.attrs)
                    except:
                        logger.info('Could not parse CF grid_mapping')

            standard_name = var.attrs['standard_name'] if 'standard_name' in var.attrs else ''
            long_name = var.attrs['long_name'] if 'long_name' in var.attrs else ''
            axis = var.attrs['axis'] if 'axis' in var.attrs else ''
            units = var.attrs['units'] if 'units' in var.attrs else ''
            CoordinateAxisType = var.attrs['CoordinateAxisType'] if 'CoordinateAxisType' in var.attrs else ''
            if standard_name == 'longitude' or \
                    CoordinateAxisType.lower() == 'lon' or \
                    long_name.lower() == 'longitude' or \
                    var_name.lower() in ['longitude', 'lon']:
                lon_var_name = var_name
            if standard_name == 'latitude' or \
                    CoordinateAxisType.lower() == 'lat' or \
                    long_name.lower() == 'latitude' or \
                    var_name.lower() in ['latitude', 'lat']:
                lat_var_name = var_name
            if (axis == 'X' or standard_name == 'projection_x_coordinate') \
                    and var.ndim == 1:
                self.xname = var_name
                # Fix for units; should ideally use udunits package
                if units == 'km':
                    self.unitfactor = 1000
                elif units == '100  km':
                    self.unitfactor = 100000
                var_data = var.values
                x = var_data*self.unitfactor
            if (axis == 'Y' or standard_name == 'projection_y_coordinate') \
                    and var.ndim == 1:
                self.yname = var_name
                # Fix for units; should ideally use udunits package
                if units == 'km':
                    self.unitfactor = 1000
                elif units == '100  km':
                    self.unitfactor = 100000
                var_data = var.values
                y = var_data*self.unitfactor
            if standard_name == 'depth' or axis == 'Z':
                var_data = var.values
                if var_data.ndim == 1:
                    if 'positive' not in var.attrs or \
                            var.attrs['positive'] == 'up':
                        self.z = var_data
                    else:
                        self.z = -var_data
            if standard_name == 'time' or axis == 'T' or var_name in ['time', 'vtime']:
                # Read and store time coverage (of this particular file)
                var_data = var.values
                time = var_data
                time_units = units

                if isinstance(time[0], np.bytes_):
                    # This hack is probably only necessary for CERSAT/GELOBCURRENT
                    time = [t.decode('ascii') for t in time]
                    self.times = [datetime.fromisoformat(t.replace('Z', '')) for t in time]
                elif time.ndim == 2:
                    self.times = [datetime.fromisoformat(''.join(t).replace('Z', '')) for t in time.astype(str)]
                else:
                    if 'calendar' in var.attrs:
                        calendar = var.attrs['calendar']
                    else:
                        calendar = 'standard'
                    self.times = num2date(time, time_units, calendar=calendar)
                self.start_time = self.times[0]
                self.end_time = self.times[-1]
                if len(self.times) > 1:
                    self.time_step = self.times[1] - self.times[0]
                else:
                    self.time_step = None
            if standard_name == 'realization':
                if ensemble_member == None:
                    var_data = var.values
                    self.realizations = var_data
                    logger.debug('%i ensemble members available'
                                % len(self.realizations))

        # Temporary workaround for Barents EPS model
        if self.realizations is None and 'ensemble_member' in self.Dataset.dims:
            self.realizations = np.arange(self.Dataset.dims['ensemble_member'])

        #########################
        # Summary of geolocation
        #########################
        if 'x' in locals():
            if x[1]-x[0] == 1 and y[1]-y[0] == 1:
                logger.info('deltaX and deltaY are 1, interpreting dataset as unprojected')
                projected = False
            else:
                projected = True
        if 'x' not in locals() or projected is False:  # No x/y-coordinates were detected
            if lon_var_name is None:
                raise ValueError('No geospatial coordinates were detected, cannot geolocate dataset')
            # We load lon and lat arrays into memory
            lon_var = self.Dataset.variables[lon_var_name]
            lat_var = self.Dataset.variables[lat_var_name]
            if lon_var.ndim == 1:
                logger.debug('Lon and lat are 1D arrays - using as projection coordinates')
                x = lon_var.data
                y = lat_var.data
                self.xname = lon_var_name
                self.yname = lat_var_name
                if self.proj4 is None:
                    self.proj4 = '+proj=latlong'
            elif lon_var.ndim == 2:
                logger.debug('Lon and lat are 2D arrays - dataset is unprojected')
                self.lon = lon_var.data
                self.lat = lat_var.data
                self.projected = False
            elif lon_var.ndim == 3:
                logger.debug('Lon lat are 3D arrays, reading first time')
                self.lon = lon_var[0,:,:].data
                self.lat = lat_var[0,:,:].data
                self.projected = False
        else:
            if self.proj4 is None:
                logger.info('Grid coordinates are detected, but proj4 string not given: assuming latlong')
                self.proj4 = '+proj=latlong'

        if 'x' in locals() and 'y' in locals():
            self.numx = len(x)
            self.numy = len(y)
            self.xmin, self.xmax = x.min(), x.max()
            self.ymin, self.ymax = y.min(), y.max()
            self.delta_x = np.abs(x[1] - x[0])
            self.delta_y = np.abs(y[1] - y[0])
            rel_delta_x = (x[1::] - x[0:-1])
            rel_delta_x = np.abs((rel_delta_x.max() -
                                  rel_delta_x.min())/self.delta_x)
            rel_delta_y = (y[1::] - y[0:-1])
            rel_delta_y = np.abs((rel_delta_y.max() -
                                  rel_delta_y.min())/self.delta_y)
            if rel_delta_x > 0.05:  # Allow 5 % deviation
                print(rel_delta_x)
                print(x[1::] - x[0:-1])
                raise ValueError('delta_x is not constant!')
            if rel_delta_y > 0.05:
                print(rel_delta_y)
                print(y[1::] - y[0:-1])
                raise ValueError('delta_y is not constant!')
            self.x = x  # Store coordinate vectors
            self.y = y

        if self.proj4 is not None and 'latlong' in self.proj4 and self.xmax is not None and self.xmax > 360:
            logger.info('Longitudes > 360 degrees, subtracting 360')
            self.xmin -= 360
            self.xmax -= 360
            self.x -= 360

        ##########################################
        # Find all variables having standard_name
        ##########################################
        self.variable_mapping = {}
        skipvars = []
        for var_name in self.Dataset.variables:
            var = self.Dataset.variables[var_name]
            if var.ndim < 2:
                continue
            if var_name in standard_name_mapping:
                # User may specify mapping if standard_name is missing, or to override existing
                standard_name = standard_name_mapping[var_name]
                self.variable_mapping[standard_name] = str(var_name)
            elif 'standard_name' in var.attrs:
                standard_name = str(var.attrs['standard_name'])
                if standard_name in self.variable_aliases:  # Mapping if needed
                    standard_name = self.variable_aliases[standard_name]
                self.variable_mapping[standard_name] = str(var_name)
            else:
                skipvars.append(var_name)

        if len(skipvars) > 0:
            logger.debug('Skipped variables without standard_name: %s' % skipvars)

        self.variables = list(self.variable_mapping.keys())

        # Run constructor of parent Reader class
        super().__init__()

    def get_variables(self, requested_variables, time=None,
                      x=None, y=None, z=None,
                      indrealization=None):

        requested_variables, time, x, y, z, _outside = self.check_arguments(
            requested_variables, time, x, y, z)

        nearestTime, dummy1, dummy2, indxTime, dummy3, dummy4 = \
            self.nearest_time(time)

        if hasattr(self, 'z') and (z is not None):
            # Find z-index range
            # NB: may need to flip if self.z is ascending
            indices = np.searchsorted(-self.z, [-z.min(), -z.max()])
            indz = np.arange(np.maximum(0, indices.min() - 1 -
                                        self.verticalbuffer),
                             np.minimum(len(self.z), indices.max() + 1 +
                                        self.verticalbuffer))
            if len(indz) == 1:
                indz = indz[0]  # Extract integer to read only one layer
        else:
            indz = 0

        if indrealization == None:
            if self.realizations is not None:
                indrealization = range(len(self.realizations))
            else:
                indrealization = None

        # Find indices corresponding to requested x and y
        if hasattr(self, 'clipped'):
            clipped = self.clipped
        else: clipped = 0

        if self.global_coverage():
            if self.lon_range() == '0to360':
                x = np.mod(x, 360)  # Shift x/lons to 0-360 
            elif self.lon_range() == '-180to180':
                x = np.mod(x + 180, 360) - 180 # Shift x/lons to -180-180
        indx = np.floor(np.abs(x-self.x[0])/self.delta_x-clipped).astype(int) + clipped
        indy = np.floor(np.abs(y-self.y[0])/self.delta_y-clipped).astype(int) + clipped
        buffer = self.buffer  # Adding buffer, to cover also future positions of elements
        indy = np.arange(np.max([0, indy.min()-buffer]),
                         np.min([indy.max()+buffer, self.numy]))
        indx = np.arange(indx.min()-buffer, indx.max()+buffer+1)

        if self.global_coverage() and indx.min() < 0 and indx.max() > 0 and indx.max() < self.numx:
            logger.debug('Requested data block is not continuous in file'+
                          ', must read two blocks and concatenate.')
            indx_left = indx[indx<0] + self.numx  # Shift to positive indices
            indx_right = indx[indx>=0]
            if indx_right.max() >= indx_left.min():  # Avoid overlap
                indx_right = np.arange(indx_right.min(), indx_left.min())
            continuous = False
        else:
            continuous = True
            indx = np.arange(np.max([0, indx.min()]),
                             np.min([indx.max(), self.numx]))

        variables = {}

        for par in requested_variables:
            if hasattr(self, 'rotate_mapping') and par in self.rotate_mapping:
                logger.debug('Using %s to retrieve %s' %
                    (self.rotate_mapping[par], par))
                if par not in self.variable_mapping:
                    self.variable_mapping[par] = \
                        self.variable_mapping[
                            self.rotate_mapping[par]]
            var = self.Dataset.variables[self.variable_mapping[par]]

            ensemble_dim = None
            if continuous is True:
                if var.ndim == 2:
                    variables[par] = var[indy, indx]
                elif var.ndim == 3:
                    variables[par] = var[indxTime, indy, indx]
                elif var.ndim == 4:
                    variables[par] = var[indxTime, indz, indy, indx]
                elif var.ndim == 5:  # Ensemble data
                    variables[par] = var[indxTime, indz, indrealization, indy, indx]
                    ensemble_dim = 0  # Hardcoded ensemble dimension for now
                else:
                    raise Exception('Wrong dimension of variable: ' +
                                    self.variable_mapping[par])
            else:  # We need to read left and right parts separately
                if var.ndim == 2:
                    left = var[indy, indx_left]
                    right = var[indy, indx_right]
                    variables[par] = np.ma.concatenate((left, right), 1)
                elif var.ndim == 3:
                    left = var[indxTime, indy, indx_left]
                    right = var[indxTime, indy, indx_right]
                    variables[par] = np.ma.concatenate((left, right), 1)
                elif var.ndim == 4:
                    left = var[indxTime, indz, indy, indx_left]
                    right = var[indxTime, indz, indy, indx_right]
                    variables[par] = np.ma.concatenate((left, right), 2)
                elif var.ndim == 5:  # Ensemble data
                    left = var[indxTime, indz, indrealization,
                               indy, indx_left]
                    right = var[indxTime, indz, indrealization,
                                indy, indx_right]
                    variables[par] = np.ma.concatenate((left, right), 3)

            variables[par] = np.asarray(variables[par])

            # Mask values outside domain
            variables[par] = np.ma.array(variables[par],
                                         ndmin=2, mask=False)
            # Mask extreme values which might have slipped through
            with np.errstate(invalid='ignore'):
                variables[par] = np.ma.masked_outside(
                    variables[par], -30000, 30000)

            # Ensemble blocks are split into lists
            if ensemble_dim is not None:
                num_ensembles = variables[par].shape[ensemble_dim]
                logger.debug('Num ensembles: %i ' % num_ensembles)
                newvar = [0]*num_ensembles
                for ensemble_num in range(num_ensembles):
                    newvar[ensemble_num] = \
                        np.take(variables[par],
                                ensemble_num, ensemble_dim)
                variables[par] = newvar

        # Store coordinates of returned points
        try:
            variables['z'] = self.z[indz]
        except:
            variables['z'] = None
        if self.projected is True:
            variables['x'] = \
                self.Dataset.variables[self.xname][indx]*self.unitfactor
            variables['y'] = \
                self.Dataset.variables[self.yname][indy]*self.unitfactor
        else:
            variables['x'] = indx
            variables['y'] = indy
        variables['x'] = np.asarray(variables['x'], dtype=np.float32)
        variables['y'] = np.asarray(variables['y'], dtype=np.float32)

        variables['time'] = nearestTime

        # Rotate any east/north vectors if necessary
        if hasattr(self, 'rotate_mapping'):
            if self.y_is_north() is True:
                logger.debug('North is up, no rotation necessary')
            else:
                self.rotate_variable_dict(variables)

        if hasattr(self, 'shift_x'):
            # "hidden feature": if reader.shift_x and reader.shift_y are defined,
            # the returned fields are shifted this many meters in the x- and y directions
            # E.g. reader.shift_x=10000 gives a shift 10 km eastwards (if x is east direction)
            if self.proj.crs.is_geographic:  # meters to degrees
                shift_y = (self.shift_y/111000)
                shift_x = (self.shift_x/111000)*np.cos(np.radians(variables['y']))
                logger.info('Shifting x between %s and %s' % (shift_x.min(), shift_x.max()))
                logger.info('Shifting y with %s m' % shift_y)
            else:
                shift_x = self.shift_x
                shift_y = self.shift_y
                logger.info('Shifting x with %s m' % shift_x)
                logger.info('Shifting y with %s m' % shift_y)
            variables['x'] += shift_x
            variables['y'] += shift_y

        return variables



#!/usr/bin/env python
#
# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Knut-Frode Dagestad, MET Norway

import sys
import argparse
import numpy as np
sys.path.append("..")

import opendrift

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('filename',
                        help='<OpenDrift output filename (netCDF)>')
    parser.add_argument('-b', dest='buffer',
                        default=0.1,
                        help='Buffer around plot in degrees lon/lat.')
    parser.add_argument('-f', dest='outfile',
                        default=None, help='Save plot to filename.')

    args = parser.parse_args()


    o = opendrift.open(args.filename)
    print(o)
    o.plot(buffer=float(args.buffer), filename=args.outfile)

if __name__ == '__main__':
    main()



# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2015, Magne Simonsen, MET Norway

import numpy as np
import logging; logger = logging.getLogger(__name__)

from opendrift.models.oceandrift import OceanDrift, Lagrangian3DArray
import pyproj

# Defining the radionuclide element properties
class Radionuclide(Lagrangian3DArray):
    """Extending Lagrangian3DArray with specific properties for radionuclides
    """

    variables = Lagrangian3DArray.add_variables([
        ('diameter', {'dtype': np.float32,
                      'units': 'm',
                      'default': 0.}),
        ('neutral_buoyancy_salinity', {'dtype': np.float32,
                                       'units': '[]',
                                       'default': 31.25}),  # for NEA Cod
        ('density', {'dtype': np.float32,
                     'units': 'kg/m^3',
                     'default': 2650.}),  # Mineral particles
        ('specie', {'dtype': np.int32,
                    'units': '',
                    'default': 0}),
#         ('transfer_rates1D', {'dtype':np.array(3, dtype=np.float32),
#                     'units': '1/s',
#                     'default': 0.})
        ('LMM_fraction', {'dtype':np.float32,
                          'units':'',
                          'default':0,
                          'seed':False}),
        ('particle_fraction', {'dtype':np.float32,
                          'units':'',
                          'default':0,
                          'seed':False})
        ])


class RadionuclideDrift(OceanDrift):
    """Radionuclide particle trajectory model based on the OpenDrift framework.

        Developed at MET Norway

        Generic module for particles that are subject to vertical turbulent
        mixing with the possibility for positive or negative buoyancy

        Particles could be e.g. oil droplets, plankton, or sediments

        Radionuclide functionality include interactions with solid matter
        (particles and sediments) through transformation processes, implemented
        with stochastic approach for speciation.

        Under construction.
    """

    ElementType = Radionuclide

    required_variables = {
        'x_sea_water_velocity': {'fallback': None},
        'y_sea_water_velocity': {'fallback': None},
        'x_wind': {'fallback': 0},
        'y_wind': {'fallback': 0},
        'land_binary_mask': {'fallback': None},
        'sea_floor_depth_below_sea_level': {'fallback': None},
        'ocean_vertical_diffusivity': {'fallback': 0.0001, 'profiles': True},
        'ocean_mixed_layer_thickness': {'fallback': 50},
        'sea_water_temperature': {'fallback': 10, 'profiles': True},
        'sea_water_salinity': {'fallback': 34, 'profiles': True},
#        'surface_downward_x_stress': {'fallback': 0},
#        'surface_downward_y_stress': {'fallback': 0},
#        'turbulent_kinetic_energy': {'fallback': 0},
#        'turbulent_generic_length_scale': {'fallback': 0},
        'upward_sea_water_velocity': {'fallback': 0},
        'conc3': {'fallback': 1.e-3},
        }

    # The depth range (in m) which profiles shall cover
    required_profiles_z_range = [-20, 0]


    def specie_num2name(self,num):
        return self.name_species[num]

    def specie_name2num(self,name):
        num = self.name_species.index(name)
        return num

    def __init__(self, *args, **kwargs):

        # Calling general constructor of parent class
        super(RadionuclideDrift, self).__init__(*args, **kwargs)

        # TODO: descriptions and units must be added in config setting below
        self._add_config({
            'radionuclide:transfer_setup': {'type': 'enum',
                'enum': ['Sandnesfj_Al','Bokna_137Cs', '137Cs_rev', 'custom'], 'default': '137Cs_rev',
                'level': self.CONFIG_LEVEL_ESSENTIAL, 'description': ''},
            'radionuclide:slowly_fraction': {'type': 'bool', 'default': False,
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            'radionuclide:irreversible_fraction': {'type': 'bool', 'default': False,
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            'radionuclide:dissolved_diameter': {'type': 'float', 'default': 0,
                'min': 0, 'max': 100e-6, 'units': 'm',
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            'radionuclide:particle_diameter': {'type': 'float', 'default': 5e-6,
                'min': .45e-6, 'max': 63.e-6, 'units': 'm',
                'level': self.CONFIG_LEVEL_ESSENTIAL, 'description': 'Mean particle diameter. Determines the settling velocity.'},
            'radionuclide:particle_diameter_uncertainty': {'type': 'float', 'default': 1e-7,
                'min': 0, 'max': 100e-6, 'units': 'm',
                'level': self.CONFIG_LEVEL_ESSENTIAL, 'description': 'Standard deviation of particle size distribution.'},
            'radionuclide:particlesize_distribution': {'type':'enum',
                'enum': ['normal','lognormal'], 'default':'normal',
                'level':self.CONFIG_LEVEL_ADVANCED, 
                'description':'Distribution of particle diameter around a mean value at seeding, NB: not at sorption!'},
            'seed:LMM_fraction': {'type': 'float','default': .1,
                'min': 0, 'max': 1, 'units': '1',
                'level': self.CONFIG_LEVEL_ESSENTIAL, 'description': 'Fraction of initial discharge released as LMM species'},
            'seed:particle_fraction': {'type': 'float','default': 0.9,
                'min': 0, 'max': 1, 'units': '1',
                'level': self.CONFIG_LEVEL_ESSENTIAL, 'description': 'Fraction of initial discharge released as particle species'},
            'seed:total_release': {'type': 'float','default': 100.e9,
                'min': 0, 'max': 1e36, 'units': 'Bq',
                'level': self.CONFIG_LEVEL_ESSENTIAL, 'description': 'Total release (Bq)'},
            # Species
            'radionuclide:species:LMM': {'type': 'bool', 'default': True,
                'level': self.CONFIG_LEVEL_BASIC, 'description': 'Toggle LMM species'},
            'radionuclide:species:LMMcation': {'type': 'bool', 'default': False,
                'level': self.CONFIG_LEVEL_BASIC, 'description': ''},
            'radionuclide:species:LMManion': {'type': 'bool', 'default': False,
                'level': self.CONFIG_LEVEL_BASIC, 'description': ''},
            'radionuclide:species:Colloid': {'type': 'bool', 'default': False,
                'level': self.CONFIG_LEVEL_BASIC, 'description': ''},
            'radionuclide:species:Humic_colloid': {'type': 'bool', 'default': False,
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            'radionuclide:species:Polymer': {'type': 'bool', 'default': False,
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            'radionuclide:species:Particle_reversible': {'type': 'bool', 'default': True,
                'level': self.CONFIG_LEVEL_BASIC, 'description': ''},
            'radionuclide:species:Particle_slowly_reversible': {'type': 'bool', 'default': False,
                'level': self.CONFIG_LEVEL_BASIC, 'description': ''},
            'radionuclide:species:Particle_irreversible': {'type': 'bool', 'default': False,
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            'radionuclide:species:Sediment_reversible': {'type': 'bool', 'default': True,
                'level': self.CONFIG_LEVEL_BASIC, 'description': ''},
            'radionuclide:species:Sediment_slowly_reversible': {'type': 'bool', 'default': False,
                'level': self.CONFIG_LEVEL_BASIC, 'description': ''},
            'radionuclide:species:Sediment_irreversible': {'type': 'bool', 'default': False,
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            # Transformations
            'radionuclide:transformations:Kd': {'type': 'float', 'default': 2.0,
                'min': 0, 'max': 1e9, 'units': 'm3/kg',
                'level': self.CONFIG_LEVEL_BASIC, 'description': ''},
            'radionuclide:transformations:Dc': {'type': 'float', 'default': 1.16e-5,
                'min': 0, 'max': 1e6, 'units': '',
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            'radionuclide:transformations:slow_coeff': {'type': 'float', 'default': 1.2e-7,
                'min': 0, 'max': 1e6, 'units': '',
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            # Sediment
            'radionuclide:sediment:sedmixdepth': {'type': 'float', 'default': 1,
                'min': 0, 'max': 100, 'units': 'm',
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            'radionuclide:sediment:sediment_density': {'type': 'float', 'default': 2600,
                'min': 0, 'max': 10000, 'units': 'kg/m3',
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            'radionuclide:sediment:effective_fraction': {'type': 'float', 'default': 0.9,
                'min': 0, 'max': 1, 'units': '',
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            'radionuclide:sediment:corr_factor': {'type': 'float', 'default': 0.1,
                'min': 0, 'max': 10, 'units': '',
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            'radionuclide:sediment:porosity': {'type': 'float', 'default': 0.6,
                'min': 0, 'max': 1, 'units': '',
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            'radionuclide:sediment:layer_thick': {'type': 'float', 'default': 1,
                'min': 0, 'max': 100, 'units': 'm',
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            'radionuclide:sediment:desorption_depth': {'type': 'float', 'default': 1,
                'min': 0, 'max': 100, 'units': 'm',
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            'radionuclide:sediment:desorption_depth_uncert': {'type': 'float', 'default': .5,
                'min': 0, 'max': 100, 'units': 'm',
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            'radionuclide:sediment:resuspension_depth': {'type': 'float', 'default': 1,
                'min': 0, 'max': 100, 'units': 'm',
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            'radionuclide:sediment:resuspension_depth_uncert': {'type': 'float', 'default': .5,
                'min': 0, 'max': 100, 'units': 'm',
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            'radionuclide:sediment:resuspension_critvel': {'type': 'float', 'default': .01,
                'min': 0, 'max': 1, 'units': 'm/s',
                'level': self.CONFIG_LEVEL_ADVANCED, 'description': ''},
            })





    def prepare_run(self):

        logger.info( 'Number of species: {}'.format(self.nspecies) )
        for i,sp in enumerate(self.name_species):
            logger.info( '{:>3} {}'.format( i, sp ) )


        logger.info( 'transfer setup: %s' % self.get_config('radionuclide:transfer_setup'))

        logger.info('nspecies: %s' % self.nspecies)
        logger.info('Transfer rates:\n %s' % self.transfer_rates)




    def init_species(self):
        # Initialize specie types
        if self.get_config('radionuclide:transfer_setup')=='Bokna_137Cs':
            self.set_config('radionuclide:species:LMM',True)
            self.set_config('radionuclide:species:Particle_reversible', True)
            self.set_config('radionuclide:species:Particle_slowly_reversible', True)
            self.set_config('radionuclide:species:Sediment_reversible', True)
            self.set_config('radionuclide:species:Sediment_slowly_reversible', True)
        elif self.get_config('radionuclide:transfer_setup')=='137Cs_rev':
            self.set_config('radionuclide:species:LMM',True)
            self.set_config('radionuclide:species:Particle_reversible', True)
            self.set_config('radionuclide:species:Sediment_reversible', True)
        elif self.get_config('radionuclide:transfer_setup')=='Sandnesfj_Al':
            self.set_config('radionuclide:species:LMM', False)
            self.set_config('radionuclide:species:LMMcation', True)
            self.set_config('radionuclide:species:LMManion', True)
            self.set_config('radionuclide:species:Humic_colloid', True)
            self.set_config('radionuclide:species:Polymer', True)
            self.set_config('radionuclide:species:Particle_reversible', True)
            self.set_config('radionuclide:species:Sediment_reversible', True)
        elif self.get_config('radionuclide:transfer_setup')=='custom':
            # Do nothing, species must be set manually
            pass
        else:
            logger.error('No valid transfer_setup {}'.format(self.get_config('radionuclide:transfer_setup')))


        self.name_species=[]
        if self.get_config('radionuclide:species:LMM'):
            self.name_species.append('LMM')
        if self.get_config('radionuclide:species:LMMcation'):
            self.name_species.append('LMMcation')
        if self.get_config('radionuclide:species:LMManion'):
            self.name_species.append('LMManion')
        if self.get_config('radionuclide:species:Colloid'):
            self.name_species.append('Colloid')
        if self.get_config('radionuclide:species:Humic_colloid'):
            self.name_species.append('Humic colloid')
        if self.get_config('radionuclide:species:Polymer'):
            self.name_species.append('Polymer')
        if self.get_config('radionuclide:species:Particle_reversible'):
            self.name_species.append('Particle reversible')
        if self.get_config('radionuclide:species:Particle_slowly_reversible'):
            self.name_species.append('Particle slowly reversible')
        if self.get_config('radionuclide:species:Particle_irreversible'):
            self.name_species.append('Particle irreversible')
        if self.get_config('radionuclide:species:Sediment_reversible'):
            self.name_species.append('Sediment reversible')
        if self.get_config('radionuclide:species:Sediment_slowly_reversible'):
            self.name_species.append('Sediment slowly reversible')
        if self.get_config('radionuclide:species:Sediment_irreversible'):
            self.name_species.append('Sediment irreversible')


        if self.get_config('radionuclide:species:Sediment_slowly_reversible') and \
                    self.get_config('radionuclide:species:Particle_slowly_reversible'):
            self.set_config('radionuclide:slowly_fraction', True)
        if self.get_config('radionuclide:species:Sediment_irreversible') and \
                    self.get_config('radionuclide:species:Particle_irreversible'):
            self.set_config('radionuclide:irreversible_fraction', True)


        self.nspecies      = len(self.name_species)
#         logger.info( 'Number of species: {}'.format(self.nspecies) )
#         for i,sp in enumerate(self.name_species):
#             logger.info( '{:>3} {}'.format( i, sp ) )







    def seed_elements(self, *args, **kwargs):

        self.init_species()

        self.init_transfer_rates()



        if 'number' in kwargs:
            num_elements = kwargs['number']
        else:
            num_elements = self.get_config('seed:number')



        if 'specie' in kwargs:
            print('num_elements', num_elements)
            try:
                print('len specie:',len(kwargs['specie']))
            except:
                print('specie:',kwargs['specie'])

            init_specie = np.ones(num_elements,dtype=int)
            init_specie[:] = kwargs['specie']
            kwargs['specie'] = init_specie[:]
            
                


        else:

            # Set initial speciation
            if 'particle_fraction' in kwargs:
                particle_frac = kwargs['particle_fraction']
            else:
                particle_frac = self.get_config('seed:particle_fraction')

            if 'LMM_fraction' in kwargs:
                lmm_frac = kwargs['LMM_fraction']
            else:
                lmm_frac = self.get_config('seed:LMM_fraction')

            shift = int(num_elements * (1-particle_frac))
            if not lmm_frac + particle_frac == 1.:
                logger.error('Fraction does not sum up to 1: %s' % str(lmm_frac+particle_frac) )
                logger.error('LMM fraction: %s ' % str(lmm_frac))
                logger.error( 'Particle fraction %s '% str(particle_frac) )
                raise ValueError('Illegal specie fraction combination : ' + str(lmm_frac) + ' '+ str(particle_frac) )

            init_specie = np.ones(num_elements, int)
            if self.get_config('radionuclide:transfer_setup')=='Sandnesfj_Al':
                init_specie[:shift] = self.num_lmmcation
            else:
                init_specie[:shift] = self.num_lmm
            init_specie[shift:] = self.num_prev

            kwargs['specie'] = init_specie


        logger.info('Initial speciation:')
        for i,sp in enumerate(self.name_species):
            logger.info( '{:>9} {:>3} {:24} '.format(  np.sum(init_specie==i), i, sp ) )

        # Set initial particle size according to speciation
        if 'diameter' in kwargs:
            diameter = kwargs['diameter']
        else:
            diameter = self.get_config('radionuclide:particle_diameter')

        init_diam =np.zeros(num_elements,float)
        if self.get_config('radionuclide:particlesize_distribution')=='normal':
            uncert  = self.get_config('radionuclide:particle_diameter_uncertainty')
            init_diam[init_specie==self.num_prev] = diameter
            init_diam[init_specie==self.num_prev] += np.random.normal(
                            0, uncert, sum(init_specie==self.num_prev))
        elif self.get_config('radionuclide:particlesize_distribution')=='lognormal':
            std = 0.5
            mu  = 0.
            if self.get_config('radionuclide:species:Particle_reversible'):
                init_diam[init_specie==self.num_prev] = \
                        np.random.lognormal(mu, std, size=sum(init_specie==self.num_prev) ) * diameter
            if self.get_config('radionuclide:species:Particle_slowly_reversible'):
                init_diam[init_specie==self.num_psrev] = \
                        np.random.lognormal(mu, std, size=sum(init_specie==self.num_psrev) ) * diameter
            if self.get_config('radionuclide:species:Particle_irreversible'):
                init_diam[init_specie==self.num_pirrev] = \
                        np.random.lognormal(mu, std, size=sum(init_specie==self.num_pirrev) ) * diameter
    

        init_diam[(init_specie==self.num_prev) & (init_diam<0.45e-6)] = 0.45e-6
        init_diam[(init_specie==self.num_prev) & (init_diam>63.5e-6)] = 63.5e-6
        if self.get_config('radionuclide:species:Particle_slowly_reversible'):
            init_diam[(init_specie==self.num_psrev) & (init_diam<0.45e-6)] = 0.45e-6
            init_diam[(init_specie==self.num_psrev) & (init_diam>63.5e-6)] = 63.5e-6
        if self.get_config('radionuclide:species:Particle_irreversible'):
            init_diam[(init_specie==self.num_pirrev) & (init_diam<0.45e-6)] = 0.45e-6
            init_diam[(init_specie==self.num_pirrev) & (init_diam>63.5e-6)] = 63.5e-6
        
        logger.info('Min: {:.2e}, Max: {:.2e}, Numer of particles seeded: {}, at {} distribution'.format(
            np.min(init_diam[init_diam>0.]), np.max(init_diam[init_diam>0.]), sum(init_diam>0.), 
             self.get_config('radionuclide:particlesize_distribution')))

        kwargs['diameter'] = init_diam



        super(RadionuclideDrift, self).seed_elements(*args, **kwargs)




    def init_transfer_rates(self):
        ''' Initialization of background values in the transfer rates 2D array.
        '''

        transfer_setup=self.get_config('radionuclide:transfer_setup')

#        logger.info( 'transfer setup: %s' % transfer_setup)


        self.transfer_rates = np.zeros([self.nspecies,self.nspecies])
        self.ntransformations = np.zeros([self.nspecies,self.nspecies])

        if transfer_setup == 'Bokna_137Cs':

            self.num_lmm    = self.specie_name2num('LMM')
            self.num_prev   = self.specie_name2num('Particle reversible')
            self.num_srev   = self.specie_name2num('Sediment reversible')
            self.num_psrev  = self.specie_name2num('Particle slowly reversible')
            self.num_ssrev  = self.specie_name2num('Sediment slowly reversible')


            # Values from Simonsen et al (2019a)
            Kd         = self.get_config('radionuclide:transformations:Kd')
            Dc         = self.get_config('radionuclide:transformations:Dc')
            slow_coeff = self.get_config('radionuclide:transformations:slow_coeff')
            susp_mat    = 1.e-3   # concentration of available suspended particulate matter (kg/m3)
            sedmixdepth = self.get_config('radionuclide:sediment:sedmixdepth')     # sediment mixing depth (m)
            default_density =  self.get_config('radionuclide:sediment:sediment_density') # default particle density (kg/m3)
            f           =  self.get_config('radionuclide:sediment:effective_fraction')      # fraction of effective sorbents
            phi         =  self.get_config('radionuclide:sediment:corr_factor')      # sediment correction factor
            poro        =  self.get_config('radionuclide:sediment:porosity')      # sediment porosity
            layer_thick =  self.get_config('radionuclide:sediment:layer_thick')      # thickness of seabed interaction layer (m)

            self.transfer_rates[self.num_lmm,self.num_prev] = Dc * Kd * susp_mat
            self.transfer_rates[self.num_prev,self.num_lmm] = Dc
            self.transfer_rates[self.num_lmm,self.num_srev] = \
                Dc * Kd * sedmixdepth * default_density * (1.-poro) * f * phi / layer_thick
            self.transfer_rates[self.num_srev,self.num_lmm] = Dc * phi
            self.transfer_rates[self.num_srev,self.num_ssrev] = slow_coeff
            self.transfer_rates[self.num_prev,self.num_psrev] = slow_coeff
            self.transfer_rates[self.num_ssrev,self.num_srev] = slow_coeff*.1
            self.transfer_rates[self.num_psrev,self.num_prev] = slow_coeff*.1


        elif transfer_setup == '137Cs_rev':

            self.num_lmm    = self.specie_name2num('LMM')
            self.num_prev   = self.specie_name2num('Particle reversible')
            self.num_srev   = self.specie_name2num('Sediment reversible')


            # Simpler version of Values from Simonsen et al (2019a)
            # Only consider the reversible fraction
            Kd         = self.get_config('radionuclide:transformations:Kd')
            Dc         = self.get_config('radionuclide:transformations:Dc')
            susp_mat    = 1.e-3   # concentration of available suspended particulate matter (kg/m3)
            sedmixdepth = self.get_config('radionuclide:sediment:sedmixdepth')     # sediment mixing depth (m)
            default_density =  self.get_config('radionuclide:sediment:sediment_density') # default particle density (kg/m3)
            f           =  self.get_config('radionuclide:sediment:effective_fraction')      # fraction of effective sorbents
            phi         =  self.get_config('radionuclide:sediment:corr_factor')      # sediment correction factor
            poro        =  self.get_config('radionuclide:sediment:porosity')      # sediment porosity
            layer_thick =  self.get_config('radionuclide:sediment:layer_thick')      # thickness of seabed interaction layer (m)

            self.transfer_rates[self.num_lmm,self.num_prev] = Dc * Kd * susp_mat
            self.transfer_rates[self.num_prev,self.num_lmm] = Dc
            self.transfer_rates[self.num_lmm,self.num_srev] = \
                Dc * Kd * sedmixdepth * default_density * (1.-poro) * f * phi / layer_thick
            self.transfer_rates[self.num_srev,self.num_lmm] = Dc * phi

        elif transfer_setup=='custom':
        # Set of custom values for testing/development

            self.num_lmm   = self.specie_name2num('LMM')
            if self.get_config('radionuclide:species:Colloid'):
                self.num_col = self.specie_name2num('Colloid')
            if self.get_config('radionuclide:species:Particle_reversible'):
                self.num_prev  = self.specie_name2num('Particle reversible')
            if self.get_config('radionuclide:species:Sediment_reversible'):
                self.num_srev  = self.specie_name2num('Sediment reversible')
            if self.get_config('radionuclide:slowly_fraction'):
                self.num_psrev  = self.specie_name2num('Particle slowly reversible')
                self.num_ssrev  = self.specie_name2num('Sediment slowly reversible')
            if self.get_config('radionuclide:irreversible_fraction'):
                self.num_pirrev  = self.specie_name2num('Particle irreversible')
                self.num_sirrev  = self.specie_name2num('Sediment irreversible')


            if self.get_config('radionuclide:species:Particle_reversible'):
                self.transfer_rates[self.num_lmm,self.num_prev] = 5.e-6 #*0.
                self.transfer_rates[self.num_prev,self.num_lmm] = \
                    self.get_config('radionuclide:transformations:Dc')
            if self.get_config('radionuclide:species:Sediment_reversible'):
                self.transfer_rates[self.num_lmm,self.num_srev] = 1.e-5 #*0.
                self.transfer_rates[self.num_srev,self.num_lmm] = \
                    self.get_config('radionuclide:transformations:Dc') * self.get_config('radionuclide:sediment:corr_factor')
#                self.transfer_rates[self.num_srev,self.num_lmm] = 5.e-6

            if self.get_config('radionuclide:slowly_fraction'):
                self.transfer_rates[self.num_prev,self.num_psrev] = 2.e-6
                self.transfer_rates[self.num_srev,self.num_ssrev] = 2.e-6
                self.transfer_rates[self.num_psrev,self.num_prev] = 2.e-7
                self.transfer_rates[self.num_ssrev,self.num_srev] = 2.e-7

        elif transfer_setup=='Sandnesfj_Al':
            # Use values from Simonsen et al (2019b)
            self.num_lmmanion    = self.specie_name2num('LMManion')
            self.num_lmmcation   = self.specie_name2num('LMMcation')
            self.num_humcol      = self.specie_name2num('Humic colloid')
            self.num_polymer     = self.specie_name2num('Polymer')
            self.num_prev        = self.specie_name2num('Particle reversible')
            self.num_srev        = self.specie_name2num('Sediment reversible')

            Dc         = self.get_config('radionuclide:transformations:Dc')

            self.salinity_intervals = [0,1,10,20]

            # Resize transfer rates array
            self.transfer_rates = np.zeros([len(self.salinity_intervals),self.transfer_rates.shape[0],self.transfer_rates.shape[1]])

            # Salinity interval 0-1 psu
            self.transfer_rates[0,self.num_lmmcation, self.num_humcol]    = 1.2e-5
            self.transfer_rates[0,self.num_lmmcation, self.num_prev]      = 4.e-6
            self.transfer_rates[0,self.num_humcol,    self.num_lmmcation] = .3*Dc
            self.transfer_rates[0,self.num_humcol,    self.num_prev]      = 2.e-6
            self.transfer_rates[0,self.num_prev,      self.num_lmmcation] = .3*Dc
            self.transfer_rates[0,self.num_srev,      self.num_lmmcation] = .03*Dc

            # Salinity interval 1-10 psu
            self.transfer_rates[1,self.num_lmmcation, self.num_humcol]    = 1.e-5
            self.transfer_rates[1,self.num_lmmcation, self.num_prev]      = 3.e-6
            self.transfer_rates[1,self.num_lmmcation, self.num_polymer]   = 1.2e-4
            self.transfer_rates[1,self.num_humcol,    self.num_lmmcation] = 7.*Dc
            self.transfer_rates[1,self.num_humcol,    self.num_prev]      = 4.e-6
            self.transfer_rates[1,self.num_prev,      self.num_lmmcation] = .5*Dc
            self.transfer_rates[1,self.num_srev,      self.num_lmmcation] = .05*Dc
            self.transfer_rates[1,self.num_lmmanion,  self.num_polymer]   = 5.e-6
            self.transfer_rates[1,self.num_polymer,   self.num_lmmanion]  = 12.*Dc
            self.transfer_rates[1,self.num_polymer,   self.num_prev]      = 2.4e-5

            # Salinity interval 10-20 psu
            self.transfer_rates[2,self.num_lmmcation, self.num_humcol]    = 8.e-6
            self.transfer_rates[2,self.num_lmmcation, self.num_prev]      = 2.e-6
            self.transfer_rates[2,self.num_lmmcation, self.num_polymer]   = 1.4e-4
            self.transfer_rates[2,self.num_humcol,    self.num_lmmcation] = 7.*Dc
            self.transfer_rates[2,self.num_humcol,    self.num_prev]      = 6.e-6
            self.transfer_rates[2,self.num_prev,      self.num_lmmcation] = .6*Dc
            self.transfer_rates[2,self.num_srev,      self.num_lmmcation] = .06*Dc
            self.transfer_rates[2,self.num_lmmanion,  self.num_polymer]   = 5.e-6
            self.transfer_rates[2,self.num_polymer,   self.num_lmmanion]  = 12.*Dc
            self.transfer_rates[2,self.num_polymer,   self.num_prev]      = 6.e-5

            # Salinity interval >20 psu
            self.transfer_rates[3,self.num_lmmcation, self.num_humcol]    = 6.e-6
            self.transfer_rates[3,self.num_lmmcation, self.num_prev]      = 1.8e-6
            self.transfer_rates[3,self.num_lmmcation, self.num_polymer]   = 1.5e-4
            self.transfer_rates[3,self.num_humcol,    self.num_lmmcation] = 7.*Dc
            self.transfer_rates[3,self.num_humcol,    self.num_prev]      = 1.e-5
            self.transfer_rates[3,self.num_prev,      self.num_lmmcation] = .8*Dc
            self.transfer_rates[3,self.num_srev,      self.num_lmmcation] = .08*Dc
            self.transfer_rates[3,self.num_lmmanion,  self.num_polymer]   = 5.e-6
            self.transfer_rates[3,self.num_polymer,   self.num_lmmanion]  = 12.*Dc
            self.transfer_rates[3,self.num_polymer,   self.num_prev]      = 8.e-5




        else:
            logger.ERROR('No transfer setup available')


        # Set diagonal to 0. (not possible to transform to present specie)
        if len(self.transfer_rates.shape) == 3:
            for ii in range(self.transfer_rates.shape[0]):
                np.fill_diagonal(self.transfer_rates[ii,:,:],0.)
        else:
            np.fill_diagonal(self.transfer_rates,0.)

#         self.transfer_rates[:] = 0.
#         print ('\n ###### \n IMPORTANT:: \n transfer rates are all set to zero for testing purposes! \n#### \n ')









    def update_terminal_velocity(self, Tprofiles=None,
                                 Sprofiles=None, z_index=None):
        """Calculate terminal velocity for Pelagic Egg

        according to
        S. Sundby (1983): A one-dimensional model for the vertical
        distribution of pelagic fish eggs in the mixed layer
        Deep Sea Research (30) pp. 645-661

        Method copied from ibm.f90 module of LADIM:
        Vikebo, F., S. Sundby, B. Aadlandsvik and O. Otteraa (2007),
        Fish. Oceanogr. (16) pp. 216-228
        """
        g = 9.81  # ms-2

        # Particle properties that determine settling velocity
        partsize = self.elements.diameter
        # prepare interpolation of temp, salt
        if not (Tprofiles is None and Sprofiles is None):
            if z_index is None:
                z_i = range(Tprofiles.shape[0])  # evtl. move out of loop
                # evtl. move out of loop
                z_index = interp1d(-self.environment_profiles['z'],
                                   z_i, bounds_error=False)
            zi = z_index(-self.elements.z)
            upper = np.maximum(np.floor(zi).astype(np.uint8), 0)
            lower = np.minimum(upper+1, Tprofiles.shape[0]-1)
            weight_upper = 1 - (zi - upper)

        # do interpolation of temp, salt if profiles were passed into
        # this function, if not, use reader by calling self.environment
        if Tprofiles is None:
            T0 = self.environment.sea_water_temperature
        else:
            T0 = Tprofiles[upper, range(Tprofiles.shape[1])] * \
                weight_upper + \
                Tprofiles[lower, range(Tprofiles.shape[1])] * \
                (1-weight_upper)
        if Sprofiles is None:
            S0 = self.environment.sea_water_salinity
        else:
            S0 = Sprofiles[upper, range(Sprofiles.shape[1])] * \
                weight_upper + \
                Sprofiles[lower, range(Sprofiles.shape[1])] * \
                (1-weight_upper)

        DENSw = self.sea_water_density(T=T0, S=S0)
        DENSpart = self.elements.density
        dr = DENSw-DENSpart  # density difference

        # water viscosity
        my_w = 0.001*(1.7915 - 0.0538*T0 + 0.007*(T0**(2.0)) - 0.0023*S0)
        # ~0.0014 kg m-1 s-1

        # terminal velocity for low Reynolds numbers
        W = (1.0/my_w)*(1.0/18.0)*g*partsize**2 * dr

        self.elements.terminal_velocity = W * self.elements.moving 


    def update_transfer_rates(self):
        '''Pick out the correct row from transfer_rates for each element. Modify the
        transfer rates according to local environmental conditions '''

        transfer_setup=self.get_config('radionuclide:transfer_setup')
        if transfer_setup == 'Bokna_137Cs' or \
         transfer_setup=='custom' or \
         transfer_setup=='137Cs_rev':
            self.elements.transfer_rates1D = self.transfer_rates[self.elements.specie,:]

            if self.get_config('radionuclide:species:Sediment_reversible'):
                # Only LMM radionuclides close to seabed are allowed to interact with sediments
                # minimum height/maximum depth for each particle
                Zmin = -1.*self.environment.sea_floor_depth_below_sea_level
                interaction_thick = self.get_config('radionuclide:sediment:layer_thick')      # thickness of seabed interaction layer (m)
                dist_to_seabed = self.elements.z - Zmin
                self.elements.transfer_rates1D[(self.elements.specie == self.num_lmm) &
                                 (dist_to_seabed > interaction_thick), self.num_srev] = 0.


            if self.get_config('radionuclide:species:Particle_reversible'):
                # Modify particle adsorption according to local particle concentration
                # (LMM -> reversible particles)
                kktmp = self.elements.specie == self.num_lmm
                self.elements.transfer_rates1D[kktmp, self.num_prev] = \
                            self.elements.transfer_rates1D[kktmp, self.num_prev] * \
                            self.environment.conc3[kktmp] / 1.e-3
        #                    self.environment.particle_conc[kktmp] / 1.e-3

        elif transfer_setup=='Sandnesfj_Al':
            sal = self.environment.sea_water_salinity
            sali = np.searchsorted(self.salinity_intervals, sal) - 1
            self.elements.transfer_rates1D = self.transfer_rates[sali,self.elements.specie,:]



    def update_speciation(self):
        '''Check if transformation processes shall occur
        Do transformation (change value of self.elements.specie)
        Update element properties for the transformed elements
        '''

        specie_in  = self.elements.specie.copy()    # for storage of the out speciation
        specie_out = self.elements.specie.copy()    # for storage of the out speciation
        deltat = self.time_step.seconds             # length of a time step
        phaseshift = np.array(self.num_elements_active()*[False])  # Denotes which trajectory that shall be transformed

        p = 1. - np.exp(-self.elements.transfer_rates1D*deltat)  # Probability for transformation
        psum = np.sum(p,axis=1)

        ran1=np.random.random(self.num_elements_active())

        # Transformation where ran1 < total probability for transformation
        phaseshift[ ran1 < psum ] = True

        logger.info('Number of transformations: %s' % sum(phaseshift))
        if sum(phaseshift) == 0:
            return

        ran4 = np.random.random(sum(phaseshift)) # New random number to decide which specie to end up in

        ttmp=[]  # list for storing the out specie
        # Loop through each trajectory
        for ii in range(sum(phaseshift)):
            # Compare random number to the relative probability for each transfer process
            ttmp.append(np.searchsorted(np.cumsum(p[phaseshift][ii]/psum[phaseshift][ii]),ran4[ii]))
        specie_out[phaseshift] = np.array(ttmp)


        # Set the new speciation
        self.elements.specie=specie_out

        logger.debug('old species: %s' % specie_in[phaseshift])
        logger.debug('new species: %s' % specie_out[phaseshift])


        for iin in range(self.nspecies):
            for iout in range(self.nspecies):
                self.ntransformations[iin,iout]+=sum((specie_in[phaseshift]==iin) & (specie_out[phaseshift]==iout))

        logger.debug('Number of transformations total:\n %s' % self.ntransformations )


        # Update radionuclide properties after transformations
        self.update_radionuclide_diameter(specie_in, specie_out)
        self.sorption_to_sediments(specie_in, specie_out)
        self.desorption_from_sediments(specie_in, specie_out)





    def sorption_to_sediments(self,sp_in=None,sp_out=None):
        '''Update radionuclide properties  when sorption to sediments occurs'''


        # Set z to local sea depth
        if self.get_config('radionuclide:species:LMM'):
            self.elements.z[(sp_out==self.num_srev) & (sp_in==self.num_lmm)] = \
                -1.*self.environment.sea_floor_depth_below_sea_level[(sp_out==self.num_srev) & (sp_in==self.num_lmm)]
            self.elements.moving[(sp_out==self.num_srev) & (sp_in==self.num_lmm)] = 0
        if self.get_config('radionuclide:species:LMMcation'):
            self.elements.z[(sp_out==self.num_srev) & (sp_in==self.num_lmmcation)] = \
                -1.*self.environment.sea_floor_depth_below_sea_level[(sp_out==self.num_srev) & (sp_in==self.num_lmmcation)]
            self.elements.moving[(sp_out==self.num_srev) & (sp_in==self.num_lmmcation)] = 0
        # avoid setting positive z values
        if np.nansum(self.elements.z>0):
            logger.debug('Number of elements lowered down to sea surface: %s' % np.nansum(self.elements.z>0))
        self.elements.z[self.elements.z > 0] = 0



    def desorption_from_sediments(self,sp_in=None,sp_out=None):
        '''Update radionuclide properties when desorption from sediments occurs'''

        desorption_depth = self.get_config('radionuclide:sediment:desorption_depth')
        std = self.get_config('radionuclide:sediment:desorption_depth_uncert')


        if self.get_config('radionuclide:species:LMM'):
            self.elements.z[(sp_out==self.num_lmm) & (sp_in==self.num_srev)] = \
                -1.*self.environment.sea_floor_depth_below_sea_level[(sp_out==self.num_lmm) & (sp_in==self.num_srev)] + desorption_depth
            self.elements.moving[(sp_out==self.num_lmm) & (sp_in==self.num_srev)] = 1
            if std > 0:
                logger.debug('Adding uncertainty for desorption from sediments: %s m' % std)
                self.elements.z[(sp_out==self.num_lmm) & (sp_in==self.num_srev)] += np.random.normal(
                        0, std, sum((sp_out==self.num_lmm) & (sp_in==self.num_srev)))
        if self.get_config('radionuclide:species:LMMcation'):
            self.elements.z[(sp_out==self.num_lmmcation) & (sp_in==self.num_srev)] = \
                -1.*self.environment.sea_floor_depth_below_sea_level[(sp_out==self.num_lmmcation) & (sp_in==self.num_srev)] + desorption_depth
            self.elements.moving[(sp_out==self.num_lmmcation) & (sp_in==self.num_srev)] = 1
            if std > 0:
                logger.debug('Adding uncertainty for desorption from sediments: %s m' % std)
                self.elements.z[(sp_out==self.num_lmmcation) & (sp_in==self.num_srev)] += np.random.normal(
                        0, std, sum((sp_out==self.num_lmmcation) & (sp_in==self.num_srev)))
        # avoid setting positive z values
        if np.nansum(self.elements.z>0):
            logger.debug('Number of elements lowered down to sea surface: %s' % np.nansum(self.elements.z>0))
        self.elements.z[self.elements.z > 0] = 0





    def update_radionuclide_diameter(self,sp_in=None,sp_out=None):
        '''Update the diameter of the radionuclides when specie is changed'''


        dia_part=self.get_config('radionuclide:particle_diameter')
        dia_diss=self.get_config('radionuclide:dissolved_diameter')


        # Transfer to reversible particles
        self.elements.diameter[(sp_out==self.num_prev) & (sp_in!=self.num_prev)] = dia_part
        std = self.get_config('radionuclide:particle_diameter_uncertainty')
        if std > 0:
            logger.debug('Adding uncertainty for particle diameter: %s m' % std)
            self.elements.diameter[(sp_out==self.num_prev) & (sp_in!=self.num_prev)] += np.random.normal(
                    0, std, sum((sp_out==self.num_prev) & (sp_in!=self.num_prev)))

        # Transfer to slowly reversible particles
        if self.get_config('radionuclide:slowly_fraction'):
            self.elements.diameter[(sp_out==self.num_psrev) & (sp_in!=self.num_psrev)] = dia_part
            if std > 0:
                logger.debug('Adding uncertainty for slowly rev particle diameter: %s m' % std)
                self.elements.diameter[(sp_out==self.num_psrev) & (sp_in!=self.num_psrev)] += np.random.normal(
                    0, std, sum((sp_out==self.num_psrev) & (sp_in!=self.num_psrev)))

        # Transfer to irreversible particles
        if self.get_config('radionuclide:irreversible_fraction'):
            self.elements.diameter[(sp_out==self.num_pirrev) & (sp_in!=self.num_pirrev)] = dia_part
            if std > 0:
                logger.debug('Adding uncertainty for irrev particle diameter: %s m' % std)
                self.elements.diameter[(sp_out==self.num_pirrev) & (sp_in!=self.num_pirrev)] += np.random.normal(
                    0, std, sum((sp_out==self.num_pirrev) & (sp_in!=self.num_pirrev)))

        # Transfer to LMM
        if self.get_config('radionuclide:species:LMM'):
            self.elements.diameter[(sp_out==self.num_lmm) & (sp_in!=self.num_lmm)] = dia_diss
        if self.get_config('radionuclide:species:LMManion'):
            self.elements.diameter[(sp_out==self.num_lmmanion) & (sp_in!=self.num_lmmanion)] = dia_diss
        if self.get_config('radionuclide:species:LMMcation'):
            self.elements.diameter[(sp_out==self.num_lmmcation) & (sp_in!=self.num_lmmcation)] = dia_diss

        # Transfer to colloids
        if self.get_config('radionuclide:species:Colloid'):
            self.elements.diameter[(sp_out==self.num_col) & (sp_in!=self.num_col)] = dia_diss
        if self.get_config('radionuclide:species:Humic_colloid'):
            self.elements.diameter[(sp_out==self.num_humcol) & (sp_in!=self.num_humcol)] = dia_diss
        if self.get_config('radionuclide:species:Polymer'):
            self.elements.diameter[(sp_out==self.num_polymer) & (sp_in!=self.num_polymer)] = dia_diss




    def bottom_interaction(self,Zmin=None):
        ''' Change speciation of radionuclides that reach bottom due to settling.
        particle specie -> sediment specie '''
        if not  ((self.get_config('radionuclide:species:Particle_reversible')) &
                  (self.get_config('radionuclide:species:Sediment_reversible')) or
                  (self.get_config('radionuclide:slowly_fraction')) or
                  (self.get_config('radionuclide:irreversible_fraction'))):
            return

        bottom = np.array(np.where(self.elements.z <= Zmin)[0])
        kktmp = np.array(np.where(self.elements.specie[bottom] == self.num_prev)[0])
        self.elements.specie[bottom[kktmp]] = self.num_srev
        self.ntransformations[self.num_prev,self.num_srev]+=len(kktmp)
        self.elements.moving[bottom[kktmp]] = 0
        if self.get_config('radionuclide:slowly_fraction'):
            kktmp = np.array(np.where(self.elements.specie[bottom] == self.num_psrev)[0])
            self.elements.specie[bottom[kktmp]] = self.num_ssrev
            self.ntransformations[self.num_psrev,self.num_ssrev]+=len(kktmp)
            self.elements.moving[bottom[kktmp]] = 0
        if self.get_config('radionuclide:irreversible_fraction'):
            kktmp = np.array(np.where(self.elements.specie[bottom] == self.num_pirrev)[0])
            self.elements.specie[bottom[kktmp]] = self.num_sirrev
            self.ntransformations[self.num_pirrev,self.num_sirrev]+=len(kktmp)
            self.elements.moving[bottom[kktmp]] = 0



    def resuspension(self):
        """ Simple method to estimate the resuspension of sedimented particles,
        checking whether the current speed near the bottom is above a critical velocity
        Sediment species -> Particle specie
        """
        # Exit function if particles and sediments not are present
        if not  ((self.get_config('radionuclide:species:Particle_reversible')) &
                  (self.get_config('radionuclide:species:Sediment_reversible'))):
            return

        specie_in = self.elements.specie.copy()

        critvel = self.get_config('radionuclide:sediment:resuspension_critvel')
        resusp_depth = self.get_config('radionuclide:sediment:resuspension_depth')
        std = self.get_config('radionuclide:sediment:resuspension_depth_uncert')

        Zmin = -1.*self.environment.sea_floor_depth_below_sea_level
        x_vel = self.environment.x_sea_water_velocity
        y_vel = self.environment.y_sea_water_velocity
        speed = np.sqrt(x_vel*x_vel + y_vel*y_vel)
        bottom = (self.elements.z <= Zmin)

        resusp = ( (bottom) & (speed >= critvel) )
        logger.info('Number of resuspended particles: {}'.format(np.sum(resusp)))
        self.elements.moving[resusp] = 1

        self.elements.z[resusp] = Zmin[resusp] + resusp_depth
        if std > 0:
            logger.debug('Adding uncertainty for resuspension from sediments: %s m' % std)
            self.elements.z[resusp] += np.random.normal(
                        0, std, sum(resusp))
        # avoid setting positive z values
        if np.nansum(self.elements.z>0):
            logger.debug('Number of elements lowered down to sea surface: %s' % np.nansum(self.elements.z>0))
        self.elements.z[self.elements.z > 0] = 0

        self.ntransformations[self.num_srev,self.num_prev]+=sum((resusp) & (self.elements.specie==self.num_srev))
        self.elements.specie[(resusp) & (self.elements.specie==self.num_srev)] = self.num_prev
        if self.get_config('radionuclide:slowly_fraction'):
            self.ntransformations[self.num_ssrev,self.num_psrev]+=sum((resusp) & (self.elements.specie==self.num_ssrev))
            self.elements.specie[(resusp) & (self.elements.specie==self.num_ssrev)] = self.num_psrev
        if self.get_config('radionuclide:irreversible_fraction'):
            self.ntransformations[self.num_sirrev,self.num_pirrev]+=sum((resusp) & (self.elements.specie==self.num_sirrev))
            self.elements.specie[(resusp) & (self.elements.specie==self.num_sirrev)] = self.num_pirrev

        specie_out = self.elements.specie.copy()
        self.update_radionuclide_diameter(specie_in, specie_out)






    def update(self):
        """Update positions and properties of radionuclide particles."""

        # Workaround due to conversion of datatype
        self.elements.specie = self.elements.specie.astype(np.int32)

        # Radionuclide speciation
        self.update_transfer_rates()
        self.update_speciation()


        # Turbulent Mixing
        if self.get_config('drift:vertical_mixing') is True:
            self.update_terminal_velocity()
            self.vertical_mixing()
        else:
            self.update_terminal_velocity()
            self.vertical_buoyancy()


        # Resuspension
        self.resuspension()
        logger.info('Speciation: {} {}'.format([sum(self.elements.specie==ii) for ii in range(self.nspecies)],self.name_species))



        # Horizontal advection
        self.advect_ocean_current()


        # Vertical advection
        if self.get_config('drift:vertical_advection') is True:
            self.vertical_advection()









# ################
# POSTPROCESSING


    def write_netcdf_radionuclide_density_map(self, filename, pixelsize_m='auto', zlevels=None,
                                              deltat=None,
                                              density_proj=None,
                                              llcrnrlon=None, llcrnrlat=None,
                                              urcrnrlon=None, urcrnrlat=None,
                                              activity_unit=None,
                                              time_avg_conc=False,
                                              horizontal_smoothing=False,
                                              smoothing_cells=0,
                                              ):
        '''Write netCDF file with map of radionuclide species densities and concentrations'''

        from netCDF4 import Dataset, date2num #, stringtochar

        logger.info('Postprocessing: Write density and concentration to netcdf file')

        if pixelsize_m == 'auto':
            lon, lat = self.get_lonlats()
            latspan = lat.max()-lat.min()
            pixelsize_m=30
            if latspan > .05:
                pixelsize_m = 50
            if latspan > .1:
                pixelsize_m = 300
            if latspan > .3:
                pixelsize_m = 500
            if latspan > .7:
                pixelsize_m = 1000
            if latspan > 2:
                pixelsize_m = 2000
            if latspan > 5:
                pixelsize_m = 4000


        if density_proj is None: # add default projection with equal-area property
            density_proj = pyproj.Proj('+proj=moll +ellps=WGS84 +lon_0=0.0')



        if activity_unit==None:
            activity_unit='Bq'  # default unit for radionuclides

        activity_per_element = self.get_config('seed:total_release') / self.num_elements_total()
        

        z = self.get_property('z')[0]
        if not zlevels==None:
            zlevels = np.sort(zlevels)
            z_array = np.append(np.append(-10000, zlevels) , max(0,np.nanmax(z)))
        else:
            z_array = [min(-10000,np.nanmin(z)), max(0,np.nanmax(z))]
        logger.info('z_array: {}'.format(  [str(item) for item in z_array] ) )




        #
        # H is array containing number of elements within each box defined by lon_array, lat_array and z_array

        H, lon_array, lat_array = \
            self.get_radionuclide_density_array(pixelsize_m, z_array,
                                                density_proj=density_proj,
                                                llcrnrlon=llcrnrlon, llcrnrlat=llcrnrlat,
                                                urcrnrlon=urcrnrlon, urcrnrlat=urcrnrlat
                                                )

        lon_array = (lon_array[:-1,:-1] + lon_array[1:,1:])/2
        lat_array = (lat_array[:-1,:-1] + lat_array[1:,1:])/2



        if horizontal_smoothing:
            # Compute horizontally smoother field
            logger.info('H.shape: ' + str(H.shape))
            Hsm = np.zeros_like(H)
            for zi in range(len(z_array)-1):
                for sp in range(self.nspecies):
                    for ti in range(H.shape[0]):
                        Hsm[ti,sp,zi,:,:] = self.horizontal_smooth(H[ti,sp,zi,:,:],n=smoothing_cells)



        # Convert from density to concentration
        logger.info('Activity: '+str(activity_per_element)+' '+ activity_unit+ ' per unit')

        # Compute mean depth and volume in each pixel grid cell
        pixel_mean_depth  =  self.get_pixel_mean_depth(lon_array, lat_array)


        pixel_volume = np.zeros_like(H[0,0,:,:,:])
        for zi,zz in enumerate(z_array[:-1]):
            topotmp = -pixel_mean_depth.copy()
            topotmp[np.where(topotmp < zz)] = zz
            topotmp = z_array[zi+1] - topotmp
            topotmp[np.where(topotmp < .1)] = 0.

            pixel_volume[zi,:,:] = topotmp * pixelsize_m**2


        pixel_volume[np.where(pixel_volume==0.)] = np.nan





        conc = np.zeros_like(H)
        if horizontal_smoothing:
            conc_sm = np.zeros_like(Hsm)
        for ti in range(H.shape[0]):
            for sp in range(self.nspecies):
                conc[ti,sp,:,:,:] = H[ti,sp,:,:,:] / pixel_volume * activity_per_element
                if horizontal_smoothing:
                    conc_sm[ti,sp,:,:,:] = Hsm[ti,sp,:,:,:] / pixel_volume * activity_per_element




        times = np.array( self.get_time_array()[0] )
        if time_avg_conc:
            conctmp = conc[:-1,:,:,:,:]
            cshape = conctmp.shape
            mdt =    np.mean(times[1:] - times[:-1])    # output frequency in opendrift output file
            if deltat==None:
                ndt = 1
            else:
                ndt = int( deltat / (mdt.seconds/3600.) )
            times2 = times[::ndt]
            times2 = times2[1:]
            odt = int(cshape[0]/ndt)
            logger.info ('ndt '+ str(ndt))   # number of time steps over which to average in conc file
            logger.info ('odt '+ str(odt))   # number of average slices


            # This may probably be written more efficiently!
            mean_conc = np.zeros( [odt,cshape[1],cshape[2],cshape[3],cshape[4]] )
            for ii in range(odt):
                meantmp  = np.mean(conctmp[(ii*ndt):(ii+1)*ndt,:,:,:,:],axis=0)
                mean_conc[ii,:,:,:,:] = meantmp






        nc = Dataset(filename, 'w')
        nc.createDimension('x', lon_array.shape[0])
        nc.createDimension('y', lon_array.shape[1])
        nc.createDimension('depth', len(z_array)-1)
        nc.createDimension('specie', self.nspecies)
        nc.createDimension('time', H.shape[0])

#        times = self.get_time_array()[0]
        timestr = 'seconds since 1970-01-01 00:00:00'
        nc.createVariable('time', 'f8', ('time',))
        nc.variables['time'][:] = date2num(times, timestr)
        nc.variables['time'].units = timestr
        nc.variables['time'].standard_name = 'time'

        if time_avg_conc:
            nc.createDimension('avg_time', odt)
            nc.createVariable('avg_time', 'f8', ('avg_time',))
            nc.variables['avg_time'][:] = date2num(times2, timestr) # np.arange(mean_conc.shape[0])
            nc.variables['avg_time'].units = timestr

        # Projection
        nc.createVariable('projection', 'i8')
        nc.variables['projection'].proj4 = density_proj.definition_string()

        #
        nc.createVariable('concfactor','f8')
        nc.variables['concfactor'][:] = activity_per_element
        nc.variables['concfactor'].long_name = 'Activity per unit element'
        nc.variables['concfactor'].unit = activity_unit

        # Cell size
        nc.createVariable('cell_size','f8')
        nc.variables['cell_size'][:] = pixelsize_m
        nc.variables['cell_size'].long_name = 'Length of cell'
        nc.variables['cell_size'].unit = 'm'

        nc.createVariable('smoothing_cells','i8')
        nc.variables['smoothing_cells'][:] = smoothing_cells
        nc.variables['smoothing_cells'].long_name = 'Number of cells in each direction for horizontal smoothing'
        nc.variables['smoothing_cells'].units = '1'

        # Coordinates
        nc.createVariable('lon', 'f8', ('y','x'))
        nc.createVariable('lat', 'f8', ('y','x'))
        nc.createVariable('depth', 'f8', ('depth',))
        nc.createVariable('specie', 'i4', ('specie',))
        nc.variables['lon'][:] = lon_array.T
        nc.variables['lon'].long_name = 'longitude'
        nc.variables['lon'].short_name = 'longitude'
        nc.variables['lon'].units = 'degrees_east'
        nc.variables['lat'][:] = lat_array.T
        nc.variables['lat'].long_name = 'latitude'
        nc.variables['lat'].short_name = 'latitude'
        nc.variables['lat'].units = 'degrees_north'
        nc.variables['depth'][:] = z_array[1:]
        nc.variables['specie'][:] = np.arange(self.nspecies)
        outstr = ['{}:{}'.format(isp,sp) for isp,sp in enumerate(self.name_species)]
        nc.variables['specie'].names = outstr



        # Density
        nc.createVariable('density', 'i4',
                          ('time','specie','depth','y', 'x'),fill_value=-999)
        H = np.swapaxes(H, 3, 4).astype('i4')
        H = np.ma.masked_where(H==0, H)
        nc.variables['density'][:] = H
        nc.variables['density'].long_name = 'Number of elements in grid cell'
        nc.variables['density'].grid_mapping = 'projection'
        nc.variables['density'].units = '1'


        if horizontal_smoothing:
            nc.createVariable('density_smooth', 'f8',
                              ('time','specie','depth','y', 'x'),fill_value=1.e36)
            Hsm = np.swapaxes(Hsm, 3, 4).astype('f8')
            #Hsm = np.ma.masked_where(Hsm==0, Hsm)
            nc.variables['density_smooth'][:] = Hsm
            nc.variables['density_smooth'].long_name = 'Horizontally smoothed number of elements in grid cell'
            nc.variables['density_smooth'].comment = 'Smoothed over '+str(smoothing_cells)+' grid points in all horizontal directions'



        # Radionuclide concentration, horizontally smoothed
        nc.createVariable('concentration', 'f8',
                          ('time','specie','depth','y', 'x'),fill_value=1.e36)
        conc = np.swapaxes(conc, 3, 4) #.astype('i4')
        #conc = np.ma.masked_where(conc==0, conc)
        nc.variables['concentration'][:] = conc
        nc.variables['concentration'].long_name = 'Radionuclide concentration'
        nc.variables['concentration'].grid_mapping = 'projection_lonlat'
        nc.variables['concentration'].units = activity_unit+'/m3'



        if horizontal_smoothing:
            # Radionuclide concentration, horizontally smoothed
            nc.createVariable('concentration_smooth', 'f8',
                              ('time','specie','depth','y', 'x'),fill_value=1.e36)
            conc_sm = np.swapaxes(conc_sm, 3, 4) #.astype('i4')
          #  conc_sm = np.ma.masked_where(conc_sm==0, conc_sm)
            nc.variables['concentration_smooth'][:] = conc_sm
            nc.variables['concentration_smooth'].long_name = 'Horizontally smoothed radionuclide concentration'
            nc.variables['concentration_smooth'].grid_mapping = 'projection_lonlat'
            nc.variables['concentration_smooth'].units = activity_unit+'/m3'
            nc.variables['concentration_smooth'].comment = 'Smoothed over '+str(smoothing_cells)+' grid points in all horizontal directions'



        if time_avg_conc:
            nc.createVariable('concentration_avg', 'f8',
                              ('avg_time','specie','depth','y', 'x'),fill_value=0)
            conc2 = np.swapaxes(mean_conc, 3, 4) #.astype('i4')
            conc2 = np.ma.masked_where(conc2==0, conc2)
            nc.variables['concentration_avg'][:] = conc2
            nc.variables['concentration_avg'].long_name = 'Time averaged radionuclide concentration'
            nc.variables['concentration_avg'].grid_mapping = 'projection_lonlat'
            nc.variables['concentration_avg'].units = activity_unit+'/m3'


        # Volume of boxes
        nc.createVariable('volume', 'f8',
                          ('depth','y', 'x'),fill_value=0)
        pixel_volume = np.swapaxes(pixel_volume, 1, 2) #.astype('i4')
        pixel_volume = np.ma.masked_where(pixel_volume==0, pixel_volume)
        nc.variables['volume'][:] = pixel_volume
        nc.variables['volume'].long_name = 'Volume of grid cell'
        nc.variables['volume'].grid_mapping = 'projection_lonlat'
        nc.variables['volume'].units = 'm3'


        # Topography
        nc.createVariable('topo', 'f8', ('y', 'x'),fill_value=0)
        pixel_mean_depth = np.ma.masked_where(pixel_mean_depth==0, pixel_mean_depth)
        nc.variables['topo'][:] = pixel_mean_depth.T
        nc.variables['topo'].long_name = 'Depth of grid point'
        nc.variables['topo'].grid_mapping = 'projection_lonlat'
        nc.variables['topo'].units = 'm'


        nc.close()
        logger.info('Wrote to '+filename)





    def get_radionuclide_density_array(self, pixelsize_m, z_array,
                                       density_proj=None, llcrnrlon=None,llcrnrlat=None,
                                       urcrnrlon=None,urcrnrlat=None,
                                       weight=None):
        '''
        compute a particle concentration map from particle positions
        Use user defined projection (density_proj=<proj4_string>)
        or create a lon/lat grid (density_proj=None)
        '''
        lon = self.get_property('lon')[0]
        lat = self.get_property('lat')[0]
        times = self.get_time_array()[0]

        # Redundant ??
        if density_proj is None: # add default projection with equal-area property
            density_proj = pyproj.Proj('+proj=moll +ellps=WGS84 +lon_0=0.0')


        # create a grid in the specified projection
        x,y = density_proj(lon, lat)
        if llcrnrlon is not None:
            llcrnrx,llcrnry = density_proj(llcrnrlon,llcrnrlat)
            urcrnrx,urcrnry = density_proj(urcrnrlon,urcrnrlat)
        else:
            llcrnrx,llcrnry = x.min()-pixelsize_m, y.min()-pixelsize_m
            urcrnrx,urcrnry = x.max()+pixelsize_m, y.max()+pixelsize_m

        x_array = np.arange(llcrnrx,urcrnrx, pixelsize_m)
        y_array = np.arange(llcrnry,urcrnry, pixelsize_m)
        bins=(x_array, y_array)
        outsidex, outsidey = max(x_array)*1.5, max(y_array)*1.5
        z = self.get_property('z')[0]
        if weight is not None:
            weight_array = self.get_property(weight)[0]

        status = self.get_property('status')[0]
        specie = self.get_property('specie')[0]
        Nspecies = self.nspecies
        H = np.zeros((len(times),
                      Nspecies,
                      len(z_array) - 1,
                      len(x_array) - 1,
                      len(y_array) - 1
                      ))

        for sp in range(Nspecies):
            for i in range(len(times)):
                if weight is not None:
                    weights = weight_array[i,:]
                else:
                    weights = None
                for zi in range(len(z_array)-1):
                    kktmp = ( (specie[i,:]==sp) & (z[i,:]>z_array[zi]) & (z[i,:]<=z_array[zi+1]) )
                    H[i,sp,zi,:,:], dummy, dummy = \
                        np.histogram2d(x[i,kktmp], y[i,kktmp],
                                   weights=weights, bins=bins)

        if density_proj is not None:
            Y,X = np.meshgrid(y_array, x_array)
            lon_array, lat_array = density_proj(X,Y,inverse=True)

        return H, lon_array, lat_array





    def get_pixel_mean_depth(self,lons,lats):
        from scipy import interpolate

        # Ocean model depth and lat/lon
        h_grd = self.conc_topo
        h_grd[np.isnan(h_grd)] = 0.
        nx = h_grd.shape[0]
        ny = h_grd.shape[1]

        lat_grd = self.conc_lat[:nx,:ny]
        lon_grd = self.conc_lon[:nx,:ny]

        # Interpolate topography to new grid
        h = interpolate.griddata((lon_grd.flatten(),lat_grd.flatten()), h_grd.flatten(), (lons, lats), method='linear')

        return h



    def horizontal_smooth(self, a, n=0):
        if n==0:
            num_coarse=a
            return num_coarse


        xdm=a.shape[1]
        ydm=a.shape[0]
        #msk = self.conc_mask
        b=np.zeros([ydm+2*n,xdm+2*n],dtype=int)
        b[n:-n,n:-n]=a


        num_coarse = np.zeros([ydm,xdm],dtype=float)
        smo_tmp1=np.zeros([ydm,xdm])
        #smo_msk1=np.zeros([ydm-2*n,xdm-2*n],dtype=float)
        nlayers = 0
        for ism in np.arange(-n,n+1):
            for jsm in np.arange(-n,n+1):
                smo_tmp = b[n+jsm:ydm+n+jsm, n+ism:xdm+n+ism]
                smo_tmp1+=smo_tmp
                # Must preferrably take care of land points
#                smo_msk = msk[n+jsm:ydm-n+jsm, n+ism:xdm-n+ism]
#                smo_msk1+=smo_msk
                nlayers+=1

        if n>0:
#            num_coarse[n:-n,n:-n] = smo_tmp1 / smo_msk1
            num_coarse[:,:] = smo_tmp1 / nlayers
        else:
            num_coarse = smo_tmp1
#        num_coarse = num_coarse*msk

        return num_coarse







    def gui_postproc(self):
        from os.path import expanduser
        
        logger.info('Postprocessing radionuclides')

        logger.info('Final speciation:')
        for isp,sp in enumerate(self.name_species):
            logger.info ('{:32}: {:>6}'.format(sp,sum(self.elements.specie==isp)))

        homefolder = expanduser("~")
        filename = homefolder+'/conc_radio.nc'
        
        
        self.guipp_saveconcfile(filename)
        
        zlayer   = [-1,-2]
        self.guipp_plotandsaveconc(filename=filename, 
                                         outfilename=homefolder+'/radio_plots/RadioConc', 
                                         zlayers=zlayer,
                                         specie= ['Total', 'LMM', 'Particle reversible'],
                                         )
        
        
        


    def guipp_saveconcfile(self,filename='none'):

        
        for readerName in self.readers:
            reader = self.readers[readerName]
            if 'sea_floor_depth_below_sea_level' in reader.variables:
                break
        
        
        self.conc_lon  = reader.get_variables('longitude',
                                       x=[reader.xmin,reader.xmax],
                                       y=[reader.ymin,reader.ymax])['longitude'][:]

        self.conc_lat  = reader.get_variables('latitude',
                                       x=[reader.xmin,reader.xmax],
                                       y=[reader.ymin,reader.ymax])['latitude'][:]

        self.conc_topo  = reader.get_variables('sea_floor_depth_below_sea_level',
                                       x=[reader.xmin,reader.xmax],
                                       y=[reader.ymin,reader.ymax])['sea_floor_depth_below_sea_level'][:]

        
        self.write_netcdf_radionuclide_density_map(filename, pixelsize_m=200.,
                                        zlevels=[-25, -2.],
                                        activity_unit='Bq',
                                        horizontal_smoothing=True,
                                        smoothing_cells=1,
                                        )


    def guipp_showanimationprofile(self):
        self.animation_profile(color='specie',
            vmin=0,vmax=self.nspecies-1,
            legend=[self.specie_num2name(i) for i in range(self.nspecies)],
            legend_loc =3,
#            markersize=10
            )
    





    def guipp_plotandsaveconc(self, filename=None, outfilename=None, zlayers=None, time=None, specie=None):
        import netCDF4
        from datetime import timedelta, datetime
        import cartopy.crs as ccrs
        import matplotlib.pyplot as plt
        from os.path import expanduser

        
        if zlayers==None:
            zlayers=[-1]
        
        
        logger.info('Plotting concentration for {} at layer {} at time {}'.format(specie, zlayers, time))
        
        homefolder = expanduser("~")
        outdir = homefolder+'/radio_plots'
        vcmin  = 0
        vcmaxt = self.get_config('seed:total_release') * 1.e-8
        cblabel= 'Bq/m$^3$'

        proj_pp=ccrs.PlateCarree()
        cmap = 'CMRmap_r'


        if specie==None:
            specie_arr = ['Total', 'LMM', 'Particle reversible', 'Sediment reversible']
        else:
            specie_arr = specie 
        
        for readerName in self.readers:
            reader = self.readers[readerName]
            if 'sea_floor_depth_below_sea_level' in reader.variables:
                break

        gtopo  = reader.get_variables('sea_floor_depth_below_sea_level',
                                       x=[reader.xmin,reader.xmax],
                                       y=[reader.ymin,reader.ymax])['sea_floor_depth_below_sea_level'][:]

        
        
        logger.info (' READ DATA FILE: {}'.format(filename))
        nc=netCDF4.Dataset(filename,'r')
        time       = nc.variables['time'][:]
        t_units    = nc.variables['time'].units
        #conc       = nc.variables['concentration_smooth'][:]; print('concentration_smooth')
        specie     = nc.variables['specie'][:]
        sp_names   = nc.variables['specie'].names
        lat        = nc.variables['lat'][:]
        lon        = nc.variables['lon'][:]
        depth      = nc.variables['depth'][:]
        topo       = nc.variables['topo'][:]
        #nc.close()

        
        
        Nspecies = len(specie[:])
        logger.info ('Nspecies: {}'.format(Nspecies))
        if Nspecies>1:
            specie_names = [i.split(':')[1] for i in sp_names]
        else:
            specie_names = [sp_names.split(':')[1]]
        logger.info(specie_names)




#        Ndepths = len(depth)
#        logger.info ('Ndepths: {}'.format(Ndepths) )
#        [print(item) for item in depth]
#         if Ndepths==1:
#             figsize=[5,6]
#         elif Ndepths==2:
#             figsize=[10,8]
#         elif Ndepths==3:
#             figsize=[13,8]
        
        
        ntime=len(time)
        time = netCDF4.num2date(time,units=t_units,only_use_cftime_datetimes=False,only_use_python_datetimes=True)
        
        
        for sp in specie_arr:
        
            fig, ax, crs, x, y, index_of_first, index_of_last = \
                self.set_up_map()
            
            cb=False

            if not sp=='Total':
                try:
                    spi = specie_names.index(sp)
                    logger.debug('{} {}'.format(sp, spi))
                except Exception:
                    logger.debug(sp)
                    break 
                
            
            for date_idx in np.arange(0,len(time)):
     
                idx1=date_idx
                t1 = time[idx1]

                for zlayer in zlayers:
                    if sp=='Total':
                        tmp = np.sum(nc.variables['concentration_smooth'][idx1,:,zlayer,:,:],axis=0)
                    else:
                        tmp = nc.variables['concentration_smooth'][idx1,spi,zlayer,:,:]
                    pcm=ax.pcolormesh(lon,lat,tmp,
                                      vmin=vcmin,vmax=vcmaxt,
                                      #norm=colors.LogNorm(vmin=vcmin,vmax=vcmax),
                                      cmap=cmap, zorder=-1,
                                      transform=proj_pp
                                      )
                    c1 = ax.contour(lon,lat,tmp,
                                   [1.e-6,1.e-5,1.e-4,1.e-3,1.e-2,1.e-1,1.e0,1.e1,1.e2,1.e3,1.e4],
                                        colors='darkgrey',linewidths=.6,zorder=4, 
                                         transform=proj_pp)
                    if not cb:
                        plt.colorbar(pcm,extend='both',label=cblabel)
                        cb=True
                    title=sp+' concentration, layer '+str(depth[zlayer])+' '+t1.strftime('%Y%m%d %H:%M')
                    fig.suptitle(title)
                    title = title.replace(' ','').replace('\n','_').replace(':','')
                    
                    if outfilename:
                        ofn = outfilename+'_'+sp.replace(' ','')+'L'+str(depth[zlayer])+'_'+t1.strftime('%Y%m%dT%H%M')+'.png'
                    else:
                        ofn = outdir+'/'+title+'_'+'.png'
                    plt.savefig(ofn, dpi=300,bbox_inches='tight')
                    
                    # remove contour lines and shading
                    for level in c1.collections:
                        level.remove()
                    pcm.remove()
                    logger.info(ofn)
            plt.close()
        logger.info('DONE')






# This file is part of OpenDrift.
#
# OpenDrift is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 2
#
# OpenDrift is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with OpenDrift.  If not, see <https://www.gnu.org/licenses/>.
#
# Copyright 2020, Gaute Hope, MET Norway

from opendrift.readers.basereader import BaseReader, ContinuousReader

import warnings
import pyproj
import numpy as np
import shapely.vectorized
import shapely.prepared
from shapely.geometry import box
import cartopy.crs as ccrs
import cartopy.feature as cfeature
import logging

logger = logging.getLogger(__name__)

# TODO: replace this with weakref + thread-safety
__roaring_mask__ = None

def get_mask_type():
    """
    Returns whether the landmask is provided by the opendrift_landmask_data
    package (0), or the roaring-landmask package (1).
    """
    try:
        from roaring_landmask import RoaringLandmask as _
        return 1
    except ImportError:
        return 0

def get_mask(skippoly = False, extent = None):
    """
    Returns an instance of the landmask type and landmask. The mask data is
    usually shared between threads.
    """
    global __roaring_mask__

    try:
        if __roaring_mask__ is None:
            from roaring_landmask import RoaringLandmask
            __roaring_mask__ = RoaringLandmask.new()

        mask = __roaring_mask__
        mask_type = 1

    except ImportError:
        from opendrift_landmask_data import Landmask
        mask_type = 0
        mask = Landmask(extent, skippoly)

    return mask_type, mask

def plot_land(ax, lonmin, latmin, lonmax, latmax, fast, ocean_color = 'white', land_color = cfeature.COLORS['land'], lscale = 'auto', globe=None):
    """
    Plot the landmask or the shapes from GSHHG.
    """
    def show_landmask_roaring(roaring):
        maxn = 512.
        dx = (lonmax - lonmin) / maxn
        dy = (latmax - latmin) / maxn
        dx = max(roaring.dx, dx)
        dy = max(roaring.dy, dy)

        x = np.arange(lonmin, lonmax, dx)
        y = np.arange(latmin, latmax, dy)

        yy, xx = np.meshgrid(y, x)
        img = roaring.mask.contains_many_par(xx.ravel(), yy.ravel()).reshape(yy.shape).T

        from matplotlib import colors
        cmap = colors.ListedColormap([ocean_color, land_color])
        ax.imshow(img, origin = 'lower', extent=[lonmin, lonmax, latmin, latmax],
                  zorder=0,
                    transform=ccrs.PlateCarree(globe=globe), cmap=cmap)

    def show_landmask_old(landmask):
        maxn = 512.
        dx = (lonmax - lonmin) / maxn
        dy = (latmax - latmin) / maxn
        dx = max(landmask.dx, dx)
        dy = max(landmask.dy, dy)

        x = np.array([lonmin, lonmax])
        y = np.array([latmin, latmax])
        ndx = np.int32(dx / landmask.dx)  # normalized delta
        ndy = np.int32(dy / landmask.dy)

        xm, ym = landmask.invtransform * (x, y)
        xm = xm.astype(np.int32)
        ym = ym.astype(np.int32)
        # TODO:
        # this may fail for small maps if xm[0]==xm[1] or ym[0]==ym[1]

        img = landmask.mask[ym[0]:ym[1]:ndy, xm[0]:xm[1]:ndx]

        from matplotlib import colors
        cmap = colors.ListedColormap([ocean_color, land_color])
        ax.imshow(img, origin = 'lower', extent=[lonmin, lonmax, latmin, latmax],
                  zorder=0,
                    transform=ccrs.PlateCarree(globe=globe), cmap=cmap)

    extent = [lonmin, latmin, lonmax, latmax]

    if fast:
        mask_type, mask = get_mask(True, extent)
        if mask_type == 0:
            show_landmask_old(mask)
        else:
            show_landmask_roaring(mask)
    else:
        if get_mask_type() == 0:
            logger.debug("Using existing GSHHS shapes..")
            _, mask = get_mask(False, extent)
            extent = box(lonmin, latmin, lonmax, latmax)
            extent = shapely.prepared.prep(extent)
            polys = [p for p in mask.polys.geoms if extent.intersects(p)]

            ax.add_geometries(polys,
                    ccrs.PlateCarree(globe=globe),
                    zorder=2,
                    facecolor=land_color,
                    edgecolor='black')
        else:
            logger.debug ("Adding GSHHS shapes..")
            f = cfeature.GSHHSFeature(scale=lscale, levels=[1],
                    facecolor=land_color)
            ax.add_geometries(
                    f.intersecting_geometries([lonmin, lonmax, latmin, latmax]),
                    ccrs.PlateCarree(globe=globe),
                    facecolor=land_color,
                    edgecolor='black')



class Reader(BaseReader, ContinuousReader):
    """
    The global landmask reader is based on the coastline data from
    GSHHG (https://www.ngdc.noaa.gov/mgg/shorelines/) optimized for
    checking against landmasks.

    Args:
        :param extent: minx, miny, maxx, maxy bounding box in source CRS for which to include
            geometries. Default None (all geometries).
        :type extent: array of floats, optional

        :param llcrnrlon: minx.
        :type llcrnrlon: float, optional, deprecated in favor of extent.

        :param llcrnrlat: miny.
        :type llcrnrlat: float, optional, deprecated in favor of extent.

        :param urcrnrlon: maxx.
        :type urcrnrlon: float, optional, deprecated in favor of extent.

        :param urcrnrlat: maxy.
        :type urcrnrlat: float, optional, deprecated in favor of extent.

        :param skippoly: use only rasterized landmask to determine whether points are on land.
        :type skippoly: bool, optional, default False.
    """
    name = 'global_landmask'
    variables = ['land_binary_mask']
    proj4 = None
    crs = None
    skippoly = False
    extent = None
    mask_type = 0  # 0 is opendrift_landmask_data, 1 is roaring_landmask. try not to make any code depend on this!

    def __init__(self,
                 extent=None,
                 llcrnrlon=None,
                 llcrnrlat=None,
                 urcrnrlon=None,
                 urcrnrlat=None,
                 skippoly=False):
        self.proj4 = '+proj=lonlat +ellps=WGS84'
        self.crs = pyproj.CRS(self.proj4)
        self.skippoly = skippoly

        super(Reader, self).__init__()

        # Depth
        self.z = None

        if extent and (llcrnrlat or llcrnrlon or urcrnrlat or urcrnrlon):
            raise Exception(
                "'extent' cannot be given togheter with any of 'llcrnrlon', ..'"
            )
        elif extent is None and (llcrnrlat or llcrnrlon or urcrnrlat
                                 or urcrnrlon):
            warnings.warn(
                "llcrnrlon, llcrnrlat, et. al. is deprecated for the global landmask reader. Prefer 'extent' in stead.",
                DeprecationWarning)
            extent = [llcrnrlon, llcrnrlat, urcrnrlon, urcrnrlat]

        # setup landmask
        self.mask_type, self.mask = get_mask(skippoly, extent)

        # Read and store min, max and step of x and y
        if self.mask_type == 0 and extent is not None:
            self.xmin, self.ymin, self.xmax, self.ymax = extent
        else:
            self.xmin, self.ymin = -180, -90
            self.xmax, self.ymax = 180, 90

        if self.mask_type == 0 and extent:
            extent = box(*[self.xmin, self.ymin, self.xmax, self.ymax])
            self.extent = shapely.prepared.prep(extent)

    def __on_land__(self, x, y):
        x = self.modulate_longitude(x)

        if self.mask_type == 0:
            return self.mask.contains(x,
                                      y,
                                      skippoly=self.skippoly,
                                      checkextent=False)
        else:
            x = x.astype(np.float64)
            y = y.astype(np.float64)
            return self.mask.contains_many_par(x, y)

    def get_variables(self,
                      requestedVariables,
                      time=None,
                      x=None,
                      y=None,
                      z=None):
        """
        Get binary mask of whether elements are on land or not.

        Args:
            x (deg[]): longitude (decimal degrees)
            y (deg[]): latitude (decimal degrees)
            ...

        x, y is given in reader local projection.

        Returns:
            Binary mask of point x, y on land.

        """

        self.check_arguments(requestedVariables, time, x, y, z)
        return {'land_binary_mask': self.__on_land__(x, y)}



"""
The main entry point of Advent is the :class:`simulation.Simulation` class.
"""

import logging; logger = logging.getLogger(__name__)

from .simulation import Simulation, ExplSimulation
from .readers import Reader, OpendriftReader, ConstantReader

def install_logs(level = logging.DEBUG):
  """
  Set up fanzy, colorful, logging.

  Args:

    level: Log level
  """
  import coloredlogs
  fmt = '%(asctime)s %(levelname)-7s %(name)s: %(message)s'
  fields = coloredlogs.DEFAULT_FIELD_STYLES
  fields['levelname']['color'] = 'magenta'
  coloredlogs.install(level = level, logger = logging.getLogger(), fmt = fmt, datefmt = None, field_styles = fields)







class EnumValidator:
    """
    validator for Enum values: a value that can only be one of a set

    """
    def __init__(self, valid_items, err_msg, case_insensitive=False):
        """
        :param valid_items: list of valid items -- can be anything that an `in`
                            test will work for.
        :param err_msg: The error message that should be used on failure.
                        Should be a format string that takes two parameters:
                        item and valid_items

        :param case_insensitive=False: whether you want the test to be
                                       case-insensitive.
                                       Only works for string values, of course.
        """
        if case_insensitive:
            valid_items = [item.lower() for item in valid_items]

        self.valid_items = valid_items
        self.err_msg = err_msg
        self.case_insensitive = case_insensitive

    def __call__(self, item):
        if self.case_insensitive:
            try:
                item = item.lower()
            except AttributeError:
                pass  # so non-strings will then fail the in test, but not crash.

        if item not in self.valid_items:
            return [self.err_msg.format(item, self.valid_items)]
        else:
            return []


class FloatRangeValidator:
    """
    Validator for float values that can only be a given range

    range is inclusive (<= and >=)
    """
    def __init__(self, min_value, max_value, err_msg=None):
        """
        :param min: minimum value allowed

        :param max: maximum value allowed

        :param err_msg: The error message that should be used on failure.
                        Should be a format string that takes three parameters:
                        default is:
                            "ValidationError: {} is not between {} and {}"
        """
        self.min = min_value
        self.max = max_value

        if err_msg is None:
            self.err_msg = "ValidationError: {} is not between {} and {}"
        else:
            self.err_msg = err_msg

    def __call__(self, value):
        try:
            value = float(value)
        except ValueError:
            return [self.err_msg.format(value, self.min, self.max)]

        if not self.min <= value <= self.max:
            return [self.err_msg.format(value, self.min, self.max)]
        else:
            return []

class YearValidator:
    """
    Validator for float values that can only be a given range

    range is inclusive (<= and >=)
    """
    def __init__(self, min_year, max_year, err_msg=None):
        """
        :param min: minimum year allowed

        :param max: maximum year allowed

        :param err_msg: The error message that should be used on failure.
                        Should be a format string that takes three parameters:
                        default is:
                            "ValidationError: {} is not between {} and {}"
        """
        self.min = min_year
        self.max = max_year

        if err_msg is None:
            self.err_msg = "ValidationError: {} is not a year between {} and {}"
        else:
            self.err_msg = err_msg

    def __call__(self, value):
        try:
            value = int(value)
        except ValueError:
            return [self.err_msg.format(value, self.min, self.max)]

        if not self.min <= value <= self.max:
            return [self.err_msg.format(value, self.min, self.max)]
        else:
            return []







from collections import OrderedDict
from datetime import datetime, timedelta

class Timeable:
    """
    Utility class for measuring total time spent in various steps in a class
    throughout program execution.
    """
    __timers__ = None
    __timing__ = None

    @property
    def timers(self):
        if self.__timers__ is None:
            self.__timers__ = OrderedDict()

        return self.__timers__

    @property
    def timing(self):
        if self.__timing__ is None:
            self.__timing__ = OrderedDict()

        return self.__timing__

    def timer_start(self, category):
        if category not in self.timing:
            self.timing[category] = timedelta(0)
        self.timers[category] = datetime.now()

    def timer_end(self, category):
        if self.timers[category] is not None:
            self.timing[category] += datetime.now() - self.timers[category]
        self.timers[category] = None




"""
cleanups that work with distillation data

There could be more, but for now, it sets the unit_type for the cuts.

"""
import math
from operator import itemgetter

import unit_conversion as uc

from .cleanup import Cleanup

from ....computation.physical_properties import Density


class FixCutUnitType(Cleanup):

    """
    sets the unit type for the cut fractions, based on the distillation type
    """

    ID = "002"  # maybe a better way to keep track of these?

    mapping = {None: 'concentration',
               'mass fraction': 'massfraction',
               'volume fraction': 'volumefraction',
               }


    def check(self):
        """
        checks to see if there is something to fix

        returns: flag, msg

        if nothing is needed, flag is None
        if something can be cleaned up, flag is True
        if something is wrong, but can not be cleaned up, flag is False

        fixme: -- maybe cleanup and validation should be better integrated?
        """
        needs_fixing = False
        for ss in self.oil.sub_samples:
            dist_data = ss.distillation_data
            for cut in dist_data.cuts:
                if cut.fraction.unit_type is not self.mapping[dist_data.type]:
                    needs_fixing = True
                    break
        if needs_fixing:
            return (True, f"Cleanup: {self.ID}: "
                    "distillation cuts have wrong unit_type -- can be corrected")
        else:
            return None, "distillation cuts have correct unit_type"


    def cleanup(self):
        """
        run this particular cleanup option
        """
        for ss in self.oil.sub_samples:
            dist_data = ss.distillation_data
            for cut in dist_data.cuts:
                cut.fraction.unit_type = self.mapping[dist_data.type]





"""
elements - LagrangianArray
==========================

This module defines the fundamental class 'LagrangianArray'

The LagrangianArray class holds a dictionary (OrderedDict) defining the variables (properties) of an element.

The variables are defined by:

    * name
    * dtype (typically numpy.float32)

and optionally:

    * unit (following udunits convention)
    * standard_name (to support output files compliant with CF convention)
    * default (default value, e.g. 0 for depth)

The values for these variables are given through the constructor, and are stored as named attributes of the created object:

.. code::

    from elements import LagrangianArray
    l = LagrangianArray(lon=[5, 6, 7], lat=[60, 60, 60], depth=10)
    print (l.lat)
    [ 60.  60.  60.]

The input values are arrays, where a single element is simply a special case of LagrangianArray with length 1.

New element types can be created by subclassing the LagrangianArray class and adding more variables through method `add_variables`.

Example for Larvae:

.. code::

    class Larvae(LagrangianArray):
        variables = LagrangianArray.add_variables({'length': {'dtype': np.float32, 'unit': 'mm'}})

New modules may easily be created from a template (template_elements.py).
Multiple inheritance is supported:

.. code::

    class CodLarvae(Larvae):
        "Extending Larvae with variables relevant for cod larvae."

        variables = Larvae.add_variables(
            {'CodLarvaeProperty1':
                {'dtype': np.float32}})


    class HalibutLarvae(Larvae):
        "Extending Larvae with variables relevant for halibut larvae."

        variables = Larvae.add_variables(
            {'HalibutLarvaeProperty1':
                {'dtype': np.float32}})


Example CodLarvae:

.. code::

    l = CodLarvae(lon=[5, 6, 7], lat=[60, 60, 60], CodLarvaeProperty1=[1], length=10)
    l.length
    10
"""

from opendrift.elements.elements import LagrangianArray
